import subprocess as sp
import traceback

from .debug import debug_print


# handle errors
def handle_unknown_exception(
    exception: Exception,
    identifier: str = "",
    reraise: bool = True,
):
    msg = """
    Caught Unexpected Exception in {id} command.
    Exception:
        {exception}
    Stacktrace:
        {stacktrace}
    """.format(
        id=identifier, exception=exception, stacktrace=traceback.format_exc()
    )
    debug_print(msg, verbose=True)
    if reraise:
        raise Exception(exception)
    else:
        return


def handle_value_error(
    exception: ValueError,
    exit_code: None | int = 0,
    identifier: str = "",
    reraise: bool = True,
):
    msg = """
    Caught ValueError in {id} command.
    Error:
        {error}
    Traceback:
        {stacktrace}
    """.format(
        id=identifier, error=exception, stacktrace=traceback.format_exc()
    )
    if reraise:
        raise ValueError(msg) from None
    else:
        return exit_code


def handle_called_process_error(
    exception: sp.CalledProcessError,
    identifier: str = "",
    reraise: bool = True,
):
    msg = """
    Caught CalledProcessError in {id} command.
    Error:
        {error}
    Traceback:
        {stacktrace}
    """.format(
        id=identifier, error=exception, stacktrace=traceback.format_exc()
    )
    debug_print(msg, verbose=True)
    if reraise:
        raise sp.CalledProcessError(exception)
    else:
        return



from .commandbuilder import CommandBuilder
from .improvers import ManifestImprover, ReadmeImprover
from .parsers import DotDict, Input, Manifest
from .utils import BX_MEMORY, BX_VCPUS, URLS, FileSystem



from functools import partial


class FunctionDictMixin:
    def do_functions(
        self, input_object: object | None, key: str | None = None, kind: str = "fixes"
    ) -> str:
        """Do all fixes. This may or may not involve transforming the manifest."""

        assigned_functions = getattr(self, kind, dict())

        relevant_functions = assigned_functions.get(key, []) + assigned_functions.get(
            None, []
        )

        if len(relevant_functions) == 0:
            return input_object  # edge-case if nothing happens.
        elif input_object is None:
            assert (
                len(relevant_functions) == 1
            ), "Not passing an input object to be progressively modified implies that you will only ever see the results of a single function. Providing more than one relevant function is useless in this case."
            func = relevant_functions[0]
            output_object = func()
            return output_object
        else:
            intermediate_object = input_object
            for func in relevant_functions:
                intermediate_object = func(intermediate_object)
            output_object = intermediate_object
            return output_object

    def add_function(
        self,
        function: str,
        key: str | None = None,
        kind="fixes",
        **kwargs,
    ):

        available_functions = getattr(self, "available_{}".format(kind), dict())
        assigned_functions = getattr(self, kind, dict())

        # process parameters
        if isinstance(function, str):
            if function in available_functions:
                function = available_functions.get(function)
            else:
                raise ValueError("Did not recognize function {f}".format(f=function))
        else:
            raise TypeError(
                "Provided function needs to be string and match one of the implemented {}!".format(
                    kind
                )
            )

        if len(kwargs) > 0:
            function = partial(function, **kwargs)

        # actual job
        if assigned_functions.get(key, None) is None:
            assigned_functions[key] = [function]
        else:
            assert isinstance(
                assigned_functions[key], list
            ), "Functions should be stored in collections associated with keys."
            assigned_functions[key].append(function)

        return



import gzip
import shutil
import subprocess as sp
import sys
from functools import partial
from pathlib import Path
from typing import Callable

from ..parsers import Input, Manifest


class Actions:
    @property
    def callables(self):
        return dict(
            mv_to_dp=self.mv_to_dp,
            ln_to_tmp=self.ln_to_tmp,
            given_pattern_find_matches_in_dp=self.given_pattern_find_matches_in_dp,
            given_dp_find_matches_with_pattern=self.given_dp_find_matches_with_pattern,
            conditional_unzip_gzip=self.conditional_unzip_gzip,
            paths_to_strings=self.paths_to_strings,
        )

    def mv_to_dp(
        self,
        input: Input,
        manifest: Manifest,
        value: str | Path | list[str] | list[Path],
        dp: str = "idp",
        as_string: bool = True,
    ) -> str | Path | list[str] | list[Path]:
        """
        Rename provided filepath(s) to be inside of a specified directory.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            value (str | Path | list[str] | list[Path]): Filename or list of filenames (as string or Paths) which should be altered to be inside of the specified directory.
            dp (str, optional): Directory path known in "self.fs". Provided filenames will be altered such that the modified filenames will be inside this directory. Defaults to "idp".
            as_string (bool, optional): Whether or not the results should be Paths or converted back to regular strings. Defaults to True.

        Returns:
            str | Path | list[str] | list[Path]: Altered Filename(s)/filepaths.
        """
        if dp in self.fs.keys():
            directory_path = getattr(self.fs, dp)
        elif Path(dp).exists():
            directory_path = Path(dp)
            if self.verbose:
                msg = """
                DIRECTORY PATH NOT FOUND IN SELF.FS
                -----------------------------------
                directory path {dp} 
                    - not found in self.fs (Keys: {k}),
                    - but refers directly to an existing directory. Using this.
                """.format(
                    dp=dp, k=set(self.fs.keys())
                )
                print(msg, flush=True)
        else:
            directory_path = Path()
            if self.verbose:
                msg = """
                DIRECTORY PATH NOT FOUND
                ------------------------
                directory path {dp} 
                    - not found in self.fs (Keys: {k})
                    - does not refer to an existing directory (Path(dp).exists() == False)
                As a fallback: replaced by current directory {r}
                """.format(
                    dp=dp, k=set(self.fs.keys()), r=directory_path.absolute()
                )
                print(msg, flush=True)

        def _rename(v):
            r = directory_path / v
            if self.verbose:
                msg = """

                RENAME FILE TO BE INSIDE DIRECTORY
                ----------------------------------
                Renamed file {f}
                to           {r}      
                """.format(
                    f=v, r=r
                )
                print(msg)
            if as_string:
                r = str(r)
            return r

        if isinstance(value, list):
            return [_rename(e) for e in value]
        else:
            return _rename(value)

    def ln_to_tmp(
        self,
        input: Input,
        manifest: Manifest,
        value: str | Path | list[str] | list[Path],
        as_string: bool = True,
    ) -> str | Path | list[str] | list[Path]:
        """
        Symlink provided filepath(s) inside /tmp directory.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            value (str | Path | list[str] | list[Path]): Filename or list of filenames (as string or Paths) which should be symlinked into the /tmp directory.
            as_string (bool, optional): Whether or not the results should be Paths or converted back to regular strings. Defaults to True.

        Returns:
            str | Path | list[str] | list[Path]: Filename(s)/filepaths in tmp.
        """

        def _command(fn):
            return "ln -s {fn} {tmp}/.".format(fn=fn, tmp=self.fs.tmp)

        def _symlink(v):
            # do symlink
            command = _command(v)
            sp.check_call(command, shell=True)

            # construct filename
            r = self.fs.tmp / Path(v).name

            if self.verbose:
                msg = """

                SYMLINK TO TMP
                --------------
                Symlinked file {f}
                to             {r}
                via command    {c}

                """.format(
                    c=command, f=v, r=r
                )
                print(msg)
            if as_string:
                r = str(r)
            return r

        if isinstance(value, list):
            return [_symlink(e) for e in value]
        else:
            return _symlink(value)

    def given_pattern_find_matches_in_dp(
        self,
        input: Input,
        manifest: Manifest,
        value: str,
        dp: str = "idp",
        as_string: bool = True,
        kind: str = "prefix",
    ) -> list[str] | list[Path]:
        dp = getattr(self.fs, dp, None)
        """
        Find filepaths in a specified directory that adhere to the pattern specified by `value`.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            value (str): (part of) pattern to match
            dp (str, optional): Directory path known in "self.fs", in which to look for files matching the pattern. Defaults to "idp".
            as_string (bool, optional): Whether or not the results should be Paths or converted back to regular strings. Defaults to True.
            kind (str, optional):   Kind of pattern you are providing as value. E.g. a prefix pattern yields "{value}*" as the `glob` pattern to matched. Defaults to "prefix".

        Returns:
            list[str] | list[Path]: List of matching filepaths.
        """
        return self.given_dp_find_matches_with_pattern(
            input,
            manifest,
            value=dp,
            pattern=value,
            as_string=as_string,
            kind=kind,
        )

    def given_dp_find_matches_with_pattern(
        self,
        input: Input,
        manifest: Manifest,
        value: str | Path,
        pattern: str = "*",
        as_string: bool = True,
        kind: str = "prefix",
    ) -> list[str] | list[Path]:
        """
        Find filepaths in a specified directory that adhere to the pattern specified by `pattern`.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            value (str | Path): Directory path where to look for the specified pattern.
            pattern (str, optional): Pattern to match. Defaults to "*".
            as_string (bool, optional): Whether or not the results should be Paths or converted back to regular strings. Defaults to True.
            kind (str, optional): Kind of pattern you are providing as value. E.g., a "prefix" pattern yields "{value}*" as the `glob` pattern to matched. Recognized options {"prefix", "suffix"}. If the option is not recognized, the assumption is a valid glob pattern was provided.  Defaults to "prefix".

        Returns:
            list[str] | list[Path]: List of matching filepaths.
        """

        dp = Path(value)
        assert (
            dp.exists()
        ), "This function expects its argument to be a valid directory. {dp} not found.".format(
            dp=dp
        )

        if kind in {"prefix"}:
            pattern = "{p}*".format(p=pattern)
        elif kind in {"suffix"}:
            pattern = "*{p}".format(p=pattern)
        else:
            pattern = pattern

        matches = list(Path(dp).glob(pattern))

        if self.verbose:
            msg = """
            Scanning directory {d}
            for files matching pattern {v}

            Found {n} matches:
                {m}
            """.format(
                d=dp, v=dp, n=len(matches), m=matches
            )
            print(msg)

        if as_string:
            return [str(fp) for fp in matches]
        else:
            return matches

    def conditional_unzip_gzip(
        self,
        input: Input,
        manifest: Manifest,
        value: str | Path,
        as_string: bool = True,
        use_pigz: bool = False,
        only_unzip_these_extensions: tuple[str] = (),
    ) -> str | Path:
        """
        Unzip a file if it is in fact gzipped.

        If so, unzip the file and return the filepath of the unzipped file.
        Otherwise, do nothing an return the original filepath.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            value (str | Path): Filename of filepath of the file that should be unzipped if applicable.
            as_string (bool, optional): Whether or not the results should be Paths or converted back to regular strings. Defaults to True.
            use_pigz (bool, optional): Whether or not `pigz` should be used to unzip a gzipped file. If set to True, the `pigz`tool needs to be available on the system. Defaults to False.

        Returns:
            str | Path: Filename or filepath of the unzipped file. Could be the original if it never was zipped in the first place.
        """

        def _should_be_unzipped(
            fp: str | Path, only_unzip_these_extensions: tuple[str] = ()
        ) -> bool:
            """Check whether provided file is gzipped yes/no."""

            with open(fp, "rb") as f:
                is_gzipped = f.read(2) == b"\x1f\x8b"

            extensions_provided = len(only_unzip_these_extensions) > 0

            if extensions_provided:
                extension_should_be_unzipped = any(
                    [str(fp).endswith(ext) for ext in only_unzip_these_extensions]
                )
            else:
                extension_should_be_unzipped = True  # Trivial pass

            return is_gzipped and extension_should_be_unzipped

        def _unzip_gzip_using_pigz(gzip_fp: Path, unzip_fp: Path):
            """Unzips a gzipped file properly using pigz."""
            if self.verbose:
                msg = """Unzipping file {fp} using pigz.
                """.format(
                    fp=gzip_fp
                )
                print(msg, flush=True)
            with open(unzip_fp, "w") as f:
                cmd = ["pigz", "-dc", str(gzip_fp)]
                exit_code = sp.check_call(cmd, stdout=f)
                if exit_code != 0:
                    sys.exit(exit_code)
            return unzip_fp

        def _unzip_gzip_using_gzip(gzip_fp: Path, unzip_fp: Path):
            """Unzips a gzipped file properly using gzip python package."""
            if self.verbose:
                msg = """Unzipping file {fp} using gzip python package.
                """.format(
                    fp=gzip_fp
                )
                print(msg, flush=True)
            with gzip.open(gzip_fp, "rb") as f_in:
                with open(unzip_fp, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
            return unzip_fp

        fp = Path(value)
        if _should_be_unzipped(
            fp, only_unzip_these_extensions=only_unzip_these_extensions
        ):
            unzip_fp = fp.parent / fp.stem  # removes .gz suffix
            if use_pigz:
                result = _unzip_gzip_using_pigz(fp, unzip_fp)
            else:
                result = _unzip_gzip_using_gzip(fp, unzip_fp)
        else:
            result = fp

        if as_string:
            return str(result)
        else:
            return result

    def paths_to_strings(
        input: Input, manifest: Manifest, value: list[Path | str]
    ) -> list[str]:
        """
        Convert a list of objects to a list of strings.

        Useful if the list contains (or could contain) Path objects, as those cannot be serialized to json, or
        inject into a shell command.

        Args:
            input (Input): Input object.
            manifest (Manifest): Manifest object.
            list_of_paths (list): List of filenames or filepaths.

        Returns:
            list[str]: List of filenames. As in: all of the entries will be strings.
        """
        return [str(p) for p in value]


class ActionMixin(Actions):
    def do_actions(self, input, manifest):
        """Do all actions. This may or may not involve transforming the input."""

        for key in self.actions.keys():
            self.do_actions_for_one_key(input, manifest, key)
        return input

    def do_actions_for_one_key(self, input, manifest, key):
        relevant_actions = self.actions.get(key, [])
        for action in relevant_actions:
            if self.verbose:
                print(
                    "Executing action for key: {k}".format(k=key),
                    flush=True,
                )

            old_val = getattr(input, key, None)  # really needs to be *inside* the loop

            if old_val is not None:
                new_val = action(input, manifest, old_val)

                if new_val is None and self.verbose:
                    print(
                        "NB: Action removed the value for key {k}.".format(k=key),
                        flush=True,
                    )

                setattr(input, key, new_val)
            else:
                """
                Here, there is no distinction between
                - an absent key (i.e. getattr(input, key, None) is None)
                - a key with value None

                Both are interpreted as "this attribute does not exist". Therefore,
                since an action represents a transformation of an attribute, no action
                is executed.
                """
                pass
        return

    def add_action(
        self, action: Callable | str = None, key: str | None = None, **kwargs
    ):
        if key is None:
            key = self.active_key

        if isinstance(action, str):
            if action in self.callables:
                action = self.callables.get(action)
            else:
                raise ValueError("Did not recognize action {a}".format(a=action))
        elif isinstance(action, Callable):
            action = action
        else:
            raise ValueError("Action needs to be string or callable.")

        # partially initialize function if desired
        if len(kwargs) > 0:
            action = partial(action, **kwargs)

        if self.actions.get(key, None) is not None:
            self.actions[key].append(action)
        else:
            self.actions[key] = [action]

        self.active_key = key
        return self



"""
Utils that are useful in dev notebooks when building an image.
"""
import json
import pprint
from pathlib import Path

from ..cli.bx_readme import run_bx_readme
from .debug import debug_print
from .run import run_command
from .io import download_resource, save_json, load_json


# datastructs
def get_fs(
    rdp: Path | None = None, tool: str = "biotool", verbose: bool = True
) -> dict:
    # fix root directory
    if rdp is None:
        rdp = Path().absolute().parent

    debug_print("Root directory path", rdp, verbose=verbose)

    # fix other directories
    fs = dict(
        rdp=rdp,
        ddp=rdp / "dev",
        mdp=rdp / "manifest",
        adp=rdp / "dev" / "assets",
        bdp=Path("bx://test/") / tool,
    )

    if fs.get("adp").exists():
        pass
    else:
        print("/assets subdirectory does not exist, making it for you.")
        fs.get("adp").mkdir(parents=True, exist_ok=True)

    # fix filepaths
    fs["mffp"] = fs.get("mdp") / "manifest.json"
    fs["rmfp"] = fs.get("mdp") / "readme.md"
    fs["dockerfile_fp"] = fs.get("rdp") / "Dockerfile"
    fs["dockerdemo_fp"] = fs.get("adp") / "Dockerdemo"
    fs["ifn"] = fs.get("adp") / "input.json"

    debug_print("Filesystem", fs, verbose=verbose)
    return fs


def get_resources(fs: dict, remote_resources: dict, verbose: bool = True) -> dict:
    local_resources = {
        k: (fs.get("adp") / Path(v).name) for k, v in remote_resources.items()
    }
    relative_local_resources = {
        k: v.relative_to(fs.get("rdp")) for k, v in local_resources.items()
    }

    bx_resources = {
        k: str(fs.get("bdp") / Path(v).name).replace("bx:/", "bx://")
        for k, v in remote_resources.items()
    }

    resources = dict(
        remote=remote_resources,
        local=local_resources,
        relative_local=relative_local_resources,
        bx=bx_resources,
    )

    debug_print("Resources", resources, verbose=verbose)

    return resources


def dl_resources(resources: dict, verbose: bool = True):
    debug_print("\n\nStarting download(s) of local resources", None, verbose=verbose)

    for n, u in resources.get("remote").items():
        f = resources.get("local").get(n)
        msg = """
        Obtaining resource {n}
        FROM: {u}
        TO:   {f}
        """.format(
            n=n, u=u, f=f
        )
        debug_print(msg, None, verbose=verbose)

        download_resource(url=u, fp=f)
    return


# commands
def get_docker_build_command(
    dockerfile_fp: Path = Path("Dockerfile"),
    tag: str = "biotool",
    cwd: Path | None = None,
    verbose: bool = True,
):
    # fix cwd
    if cwd is None:
        cwd = Path().absolute().parent
        debug_print("cwd", cwd, verbose=verbose)

    # build command
    cmd = "docker build . -f {f} -t {t}".format(f=dockerfile_fp.relative_to(cwd), t=tag)

    debug_print(
        "docker build command",
        cmd,
        verbose=verbose,
    )
    return cmd


def get_bx_run_local_command(
    input_dict_fp: Path = Path("input.json"),
    tag: str = "tool",
    cwd: Path | None = None,
    short: bool = True,
    verbose: bool = True,
):
    # fix cwd
    if cwd is None:
        cwd = Path().absolute().parent
        debug_print("cwd", cwd, verbose=verbose)

    if short:
        cmd_input_part = "$(cat {fp})".format(fp=input_dict_fp.relative_to(cwd))
    else:
        input_dict = load_json(input_dict_fp)
        cmd_input_part = "{ip}".format(ip=input_dict).replace("'", '"')

    cmd = "bx run-local {t} '{input_part}'".format(t=tag, input_part=cmd_input_part)

    debug_print(
        "bx run-local command",
        cmd,
        verbose=verbose,
    )
    return cmd


# setup of experiments
def get_and_run_docker_build_command(
    dockerfile_fp: Path = Path("Dockerfile"),
    tag: str = "biotool",
    cwd: Path | None = None,
    verbose: bool = True,
):
    # fix cwd
    if cwd is None:
        cwd = Path().absolute().parent
        debug_print("cwd", cwd, verbose=verbose)

    docker_build_command = get_docker_build_command(
        dockerfile_fp=dockerfile_fp, tag=tag, cwd=cwd, verbose=verbose
    )
    run_command(
        docker_build_command, identifier="Docker build", cwd=cwd, verbose=verbose
    )
    return


def setup_bx_run_local(
    resources: dict = {},
    input_dict_fp: Path = "input.json",
    tag: str = "tool",
    cwd: Path | None = None,
    short: bool = True,
    verbose: bool = True,
    **kwargs
):

    input_dict = get_input_dict(
        resources=resources,
        input_dict_fp=input_dict_fp,
        cwd=cwd,
        verbose=verbose,
        **kwargs
    )

    bx_run_local_command = get_bx_run_local_command(
        input_dict_fp=input_dict_fp, tag=tag, cwd=cwd, verbose=verbose, short=short
    )
    return input_dict, bx_run_local_command


def get_input_dict(
    resources: dict = {},
    input_dict_fp: Path | None = None,
    cwd: Path | None = None,
    verbose: bool = True,
    **kwargs
):
    def _make_paths_relative_recursively(d: dict, cwd: Path):
        for k, v in d.items():
            if isinstance(v, Path):
                d[k] = v.relative_to(cwd)
            elif isinstance(v, list | tuple | set):
                d[k] = [e.relative_to(cwd) if isinstance(v, Path) else e for e in v]
            elif isinstance(v, dict):
                d[k] = _make_paths_relative_recursively(v, cwd=cwd)
            else:
                pass
        return d

    if cwd is not None:
        resources = _make_paths_relative_recursively(resources, cwd=cwd)

    input_dict = {**resources, **kwargs}
    debug_print("input_dict", input_dict, verbose=verbose)

    if input_dict_fp is not None:
        save_json(input_dict, fp=input_dict_fp, verbose=verbose)
    return input_dict



from pathlib import Path

import jinja2
from jinja2 import Environment, FileSystemLoader, select_autoescape

from ..functiondict_mixin import FunctionDictMixin

ROOT = Path(__file__).absolute().parent.parent
TEMPLATE_DP = ROOT / "templates"


class Templates:
    # properties
    @property
    def available_templates(self):
        return {t.stem: t.name for t in self.template_dp.glob("*.jinja")}

    @property
    def active_template(self):
        return getattr(self, "_active_template", None)

    @active_template.setter
    def active_template(self, value: str | None):
        if value is None:
            if self.is_pipeline:
                value = "pipeline-inputs.jinja"
            else:
                value = "tool-inputs.jinja"
        elif value in self.available_templates:
            value = self.available_templates.get(value)
        else:
            msg = """
            Did not find template: {v}
            Available templates:   {ts}

            Template directory:    {td}
            Does it exist?         {exists}
            Contents of template directory: {tc}


            Root directory:        {rd}
            Contents of root directory: {c}
            """.format(
                v=value,
                ts=self.available_templates,
                td=TEMPLATE_DP,
                exists=TEMPLATE_DP.exists(),
                tc=list(TEMPLATE_DP.iterdir()),
                rd=ROOT,
                c=list(ROOT.iterdir()),
            )
            raise ValueError(msg)

        self._active_template = self.jinja_environment.get_template(value)
        return

    @property
    def jinja_environment(self):
        if getattr(self, "_jinja_environment", None) is None:
            self._jinja_environment = Environment(
                loader=FileSystemLoader(self.template_dp),
                autoescape=select_autoescape(),
                trim_blocks=True,
                lstrip_blocks=True,
            )
        return self._jinja_environment


class Renders(Templates):
    template_dp = TEMPLATE_DP

    @property
    def available_renders(self):
        return dict(
            title=self.title,
            context=self.context,
            examples=self.examples,
            links=self.links,
            tool_versions=self.tool_versions,
            tool_inputs=self.tool_inputs,
            tool_outputs=self.tool_outputs,
            pipeline_inputs=self.pipeline_inputs,
            pipeline_outputs=self.pipeline_outputs,
            pipeline_overview=self.pipeline_overview,
            pipeline_steps=self.pipeline_steps,
        )

    # drafts
    def init_readme(self, is_pipeline: bool = False):
        if is_pipeline:
            return self._render(template="init-readme-pipeline", d=dict())
        else:
            return self._render(template="init-readme-tool", d=dict())

    def context(self):
        return self._render(template="context", d=dict())

    def examples(self):
        return self._render(template="examples", d=dict())

    # Draft & Manifest-defined
    def links(self, manifest_dict: dict = dict()):
        return self._render(template="links", d=manifest_dict)

    def tool_versions(self, manifest_dict: dict = dict()):
        return self._render(template="tool-versions", d=manifest_dict)

    # Manifest-defined
    def title(self, manifest_dict: dict = dict()):
        return self._render(template="title", d=manifest_dict)

    def tool_inputs(self, manifest_dict: dict = dict()):
        return self._render(template="tool-inputs", d=manifest_dict)

    def tool_outputs(self, manifest_dict: dict = dict()):
        return self._render(template="tool-outputs", d=manifest_dict)

    def pipeline_inputs(self, manifest_dict: dict = dict()):
        return self._render(template="pipeline-inputs", d=manifest_dict)

    def pipeline_outputs(self, manifest_dict: dict = dict()):
        return self._render(template="pipeline-outputs", d=manifest_dict)

    def pipeline_overview(self, steps_dict: dict = dict()):
        return self._render(template="pipeline-overview", d=steps_dict)

    def pipeline_steps(self, steps_dict: dict = dict()):
        return self._render(template="pipeline-steps", d=steps_dict)

    # underlying worker
    def _render(self, template: str, d: dict, as_lines: bool = True):
        self.active_template = template
        rendered_content = self.active_template.render(**d)
        if as_lines:
            rendered_content = rendered_content.split("\n")
            rendered_content = ["{c}\n".format(c=c) for c in rendered_content]
            return rendered_content
        else:
            return rendered_content


class RendersMixin(Renders, FunctionDictMixin):
    def do_renders(
        self, manifest_dict: dict | None = None, key: str | None = None
    ) -> None:
        return self.do_functions(manifest_dict, key=key, kind="renders")

    def add_render(
        self,
        render: str,
        key: str | None = None,
        **kwargs,
    ):
        return self.add_function(function=render, key=key, kind="renders", **kwargs)

    def add_renders(
        self,
        renders: list[str],
        key: str | None = None,
        **kwargs,
    ):
        for render in renders:
            self.add_render(render, key=key, **kwargs)
        return



from ..parsers import Input, Manifest
from ..utils import FileSystem
from .actions import ActionMixin
from .constraints import ConstraintMixin
from .kinds import CommandKindMixin


class CommandBuilder(ActionMixin, ConstraintMixin, CommandKindMixin):
    def __init__(
        self,
        manifest: Manifest,
        tool: str = "toolname",
        filesystem: FileSystem | None = None,
        verbose=True,
        default_prefix="--",
        default_kind="keyval",
    ):

        self.tool = tool
        self.fs = filesystem if filesystem is not None else FileSystem(tool=tool)
        self.default_prefix = default_prefix
        self.default_kind = default_kind

        self.verbose = verbose
        self.manifest = manifest

        self.active_key = None

        self.kinds = dict(
            keyval=set([]), flag=set([]), argument=set([])
        )  # kind -> set of keys
        self.actions = dict()  # key -> list of actions
        self.commands = dict()  # key -> cmd str
        self.constraints = dict()  # key -> list of constraints
        return

    def add_command(
        self,
        cmd: str = "",
        key: str | None = None,
        prefix: str | None = None,
        kind: str | None = None,
    ):
        # assign defaults in absence of explicit values
        if key is None:
            key = self.active_key

        if prefix is None:
            prefix = self.default_prefix  # fallback on default

        if kind is None:
            kind = self.default_kind  # fallback on default

        # generate command part (i.e., a string) and add to dict
        self.commands[key] = "{prefix}{command}".format(prefix=prefix, command=cmd)

        self.active_key = key

        if kind is not None:
            return self.add_kind(kind=kind)
        else:
            return self

    def build_command(self, input: Input):
        return " ".join(self.get_cmd(input)).strip()

    # properties
    @property
    def manifest(self):
        return getattr(self, "_manifest", None)

    @manifest.setter
    def manifest(self, value):
        if isinstance(value, Manifest):
            self._manifest = value
        else:
            raise ValueError("manifest attribute only accepts Manifest object.")

    @property
    def fs(self):
        return getattr(self, "_fs", None)

    @fs.setter
    def fs(self, value):
        if isinstance(value, FileSystem):
            self._fs = value
        else:
            raise ValueError("fs attribute only accepts FileSystem object.")



from pathlib import Path

from ..parsers import Manifest
from ..parsers.api import BxImage
from .guidelines import ChecksMixin, EnrichmentsMixin, FixesMixin


class ManifestImproverInspectorMixin:
    def replace_suffix(self, jsonpath: str, suffix: str):
        parts = jsonpath.split(".")
        parts[-1] = suffix  # replace suffix
        return ".".join(parts)

    def find_all(
        self,
        suffix: list[str] | str = "title",
        prefix: list[str] | str = "",
        contains: list[str] | str = "",
    ):
        if isinstance(contains, str):
            contains = [contains]
        if isinstance(suffix, str):
            suffix = [suffix]
        if isinstance(prefix, str):
            prefix = [prefix]

        return [
            jp
            for jp in self.mf.collect_jsonpaths()
            if any((jp.startswith(s) for s in prefix))
            if any((jp.endswith(s) for s in suffix))
            if any((c in jp for c in contains))
        ]

    def find_manifest_entries_associated_with_steps(
        self,
        steps: list[str] | dict | None = None,
        contains_prefix="tools.properties",
        suffix=["description", "title"],
    ):
        if steps is None:
            steps = self.steps

        contains = [".".join([contains_prefix, s]) for s in steps]
        return self.find_all(
            contains=contains,
            suffix=suffix,
        )

    def collect_flat_dict(self, jsonpath: str, field="default"):
        def format_key(k, jp):
            k = k.split(jp)[1]
            parts = k.split(".")
            parts = [p for p in parts if p not in {"properties", field, ""}]
            return ".".join(parts)

        all_jps = self.find_all(suffix=field)
        relevant_jps = [d for d in all_jps if d.startswith(jsonpath)]
        lowest_level_relevant_jps = []
        for d in relevant_jps:
            if (
                len(
                    [
                        e
                        for e in relevant_jps
                        if e.startswith(d.split(".{f}".format(f=field))[0])
                    ]
                )
                == 1
            ):
                lowest_level_relevant_jps.append(d)

        flat_dict = {
            format_key(jp, jsonpath): self.mf[jp] for jp in lowest_level_relevant_jps
        }
        flat_dict = {k: v for k, v in flat_dict.items() if k != ""}
        return flat_dict

    def collect_defaults(self, jsonpath: str):
        return self.collect_flat_dict(jsonpath=jsonpath, field="default")

    def collect_titles(self, jsonpath: str):
        return self.collect_flat_dict(jsonpath=jsonpath, field="title")

    def collect_descendants(
        self,
        jsonpath: str,
        key_reference_field: str = "title",
        value_reference_field: str = "default",
        dummy_value=None,
    ):
        keys = self.collect_flat_dict(jsonpath=jsonpath, field=key_reference_field)
        values = self.collect_flat_dict(jsonpath=jsonpath, field=value_reference_field)
        return {k: values.get(k, dummy_value) for k in keys}

    @property
    def titles(self):
        return self.find_all(suffix="title")

    @property
    def descriptions(self):
        return self.find_all(suffix="description")

    @property
    def required(self):
        return self.find_all(suffix="required")

    @property
    def defaults(self):
        return self.find_all(suffix="default")

    @property
    def steps(self):
        if self.mf.is_pipeline:
            if getattr(self, "_steps", None) is None:
                steps = dict()
                for step in self.mf.pipeline.steps:
                    img = BxImage(step)

                    steps[img.canonical_name] = dict(
                        step=step,
                        name=img.name,
                        title=img.title,
                        env=img.environment,
                        version=img.version,
                        urls=img.urls,
                    )
                self._steps = steps
            else:
                pass

            return self._steps
        else:
            return None


class ManifestImprover(
    ChecksMixin, FixesMixin, EnrichmentsMixin, ManifestImproverInspectorMixin
):
    def __init__(
        self,
        manifest_fp: str | Path | None = None,
        manifest: Manifest | None = None,
        verbose: bool = True,
    ):
        if manifest_fp is not None:
            self.manifest = Manifest(manifest_fp, jsonpaths=True)
        elif manifest is not None:
            self.manifest = manifest
        else:
            raise ValueError("Need a manifest filepath or manifest.")

        self.fp = self.mf.fp
        self.verbose = verbose

        self.fixes = dict()
        self.checks = dict()
        self.enrichments = dict()
        return

    def do_improvements(self, line: str, key: str | None = None) -> str:
        self.do_checks(line, key=key)
        line = self.do_fixes(line, key=key)
        line = self.do_enrichments(line, key=key)
        return line

    def improve_lines(
        self,
        sources: list[str],
        targets: list[str] | None = None,
        source_values: list[str] | None = None,
        target_values: list[str] | None = None,
        key: str = "title",
        overwrite: bool = True,
    ):
        if targets is None:
            targets = sources

        if source_values is None:
            old_source_values = [getattr(self.manifest, src, None) for src in sources]
        else:
            old_source_values = source_values

        if target_values is None:
            old_target_values = [getattr(self.manifest, tgt, None) for tgt in targets]
        else:
            old_target_values = target_values

        for src, old_val_src, tgt, old_val_tgt in zip(
            sources, old_source_values, targets, old_target_values
        ):
            new_val_src = self.do_improvements(old_val_src, key=key)

            change = old_val_src != new_val_src
            if new_val_src is None:
                pass
            elif overwrite:
                setattr(self.manifest, tgt, new_val_src)
            elif change and old_val_tgt is None:
                setattr(self.manifest, tgt, new_val_src)
            else:
                pass
        return

    def improve_titles(self):
        return self.improve_lines(sources=self.titles, key="title")

    def improve_descriptions(self):
        return self.improve_lines(sources=self.descriptions, key="description")

    def improve_descriptions_from_title(self):
        sources = [t for t in self.titles if t != "title"]
        targets = [self.replace_suffix(t, suffix="description") for t in sources]
        return self.improve_lines(
            sources=sources,
            targets=targets,
            key="description_from_title",
            overwrite=False,
        )

    def improve_descriptions_from_name(self, keys: list[str] | None = None):
        sources = self.descriptions
        source_values = sources  # input is actual name

        if keys is None:
            keys = [
                k
                for k in self.enrichments
                if any(
                    [
                        s in k
                        for s in {
                            "description_from_name",
                            "descriptions_from_name",
                            "description_from_names",
                            "descriptions_from_names",
                        }
                    ]
                )
            ]

        for key in keys:
            self.improve_lines(
                sources=sources,
                source_values=source_values,
                key=key,
                overwrite=True,
            )
        return

    def improve_titles_from_name(self, keys: list[str] | None = None):
        sources = self.titles
        source_values = sources  # input is actual name

        if keys is None:
            keys = [
                k
                for k in self.enrichments
                if any(
                    [
                        s in k
                        for s in {
                            "title_from_name",
                            "titles_from_name",
                            "title_from_names",
                            "titles_from_names",
                        }
                    ]
                )
            ]

        for key in keys:
            self.improve_lines(
                sources=sources,
                source_values=source_values,
                key=key,
                overwrite=True,
            )
        return

    def improve_defaults_from_name(self, keys: list[str] | None = None):
        sources = self.defaults
        source_values = sources  # input is actual name

        if keys is None:
            keys = [
                k
                for k in self.enrichments
                if any(
                    [
                        s in k
                        for s in {
                            "default_from_name",
                            "defaults_from_name",
                            "default_from_names",
                            "defaults_from_names",
                        }
                    ]
                )
            ]

        for key in keys:
            self.improve_lines(
                sources=sources,
                source_values=source_values,
                key=key,
                overwrite=True,
            )
        return

    def improve_available_from_name(self, keys: list[str] | None = None):
        sources = [t for t in self.titles if t != "title"]
        targets = [self.replace_suffix(t, suffix="available") for t in sources]

        sources = [t.split(".title")[0] for t in self.titles if t != "title"]
        source_values = sources  # input is actual name

        if keys is None:
            keys = [
                k
                for k in self.enrichments
                if any(
                    [
                        s in k
                        for s in {
                            "available_from_name",
                            "availables_from_name",
                            "available_from_names",
                            "availables_from_names",
                        }
                    ]
                )
            ]

        for key in keys:
            self.improve_lines(
                sources=sources,
                source_values=source_values,
                targets=targets,
                key=key,
                overwrite=True,
            )
        return

    # properties
    @property
    def mf(self):
        # alias
        return self.manifest

    @mf.setter
    def mf(self, value):
        # alias
        self.manifest = value
        return

    @property
    def manifest(self):
        return getattr(self, "_manifest", None)

    @manifest.setter
    def manifest(self, value):
        if isinstance(value, Manifest):
            self._manifest = value
        else:
            raise TypeError("manifest attribute only accepts Manifest object.")



from functools import partial

from ...parsers import DotDict
from ...utils import FIXED_DESCRIPTIONS, URLS
from ..functiondict_mixin import FunctionDictMixin


class Enrichments:
    @property
    def available_enrichments(self):
        return dict(
            document_urls=self.document_urls,
            fill_in_fixed_description=self.fill_in_fixed_description,
            fetch_external_title=self.fetch_external_title,
            fetch_defaults=self.fetch_defaults,
            fetch_available=self.fetch_available,
            fetch_flat_dict_for_field=self.fetch_flat_dict_for_field,
        )

    def document_urls(
        self,
        line: str,
        urls_to_document: None | dict = None,
        verbose: bool | None = None,
    ) -> str:
        if verbose is None:
            verbose = self.verbose

        def _update_urls_to_document(
            line: str, urls_to_document: dict | None, verbose: bool
        ):
            if urls_to_document is None:
                urls_to_document = URLS.copy()
            else:
                urls_to_document.copy()

            # count already documented urls
            counts = {k: line.count(v) for k, v in urls_to_document.items()}

            for k, c in counts.items():
                if c > 0:
                    # this url has been documented already
                    urls_to_document.pop(k)

                if c > 1:
                    # this url has been excessively documented

                    if verbose:
                        msg = """
                        In the following excerpt (first 100 chars shown): 
                        '''
                        {e}
                        '''

                        Term     {k}
                        with URL {u}
                        has been documentend {n} times 

                        Please verify whether this was a conscious decision, as the 
                        documentation guidelines specify documenting the same term just once.
                        """.format(
                            k=k,
                            u=urls_to_document.get(
                                k,
                                "<URL MISSING>: It did get popped from the dict as it should.",
                            ),
                            n=c,
                            e=line[: min(100, len(line))],
                        )
                        print(msg)
            return urls_to_document

        def _add_url_to_line(line: str, key: str, url) -> str:
            attempt_01 = line.replace(
                " {k}".format(k=k), " [{k}]({u})".format(k=k, u=u), 1
            )

            attempt_02 = line.replace(
                "{k} ".format(k=k), "[{k}]({u}) ".format(k=k, u=u), 1
            )

            if attempt_01 != line:
                new_line = attempt_01
                if verbose:
                    print("Documented an instance of {}".format(k))
            elif attempt_02 != line:
                new_line = attempt_02
                if verbose:
                    print("Documented an instance of {}".format(k))
            else:
                new_line = line
            return new_line

        urls_to_document = _update_urls_to_document(line, urls_to_document, verbose)

        new_line = line
        for k, u in urls_to_document.items():
            new_line = _add_url_to_line(new_line, k, u)

        return new_line

    def fill_in_fixed_description(
        self,
        name: str,
        fixed_descriptions: None | dict = None,
        warn: bool = False,
        verbose: bool | None = None,
    ) -> str | None:

        if verbose is None:
            verbose = self.verbose

        if fixed_descriptions is None:
            fixed_descriptions = FIXED_DESCRIPTIONS.copy()

        fixed_description = fixed_descriptions.get(name, None)

        if fixed_description is not None and verbose:
            msg = """
            [ENRICHMENT] Found a description for which the guidelines specify a fixed text.
                THIS PARAMETER: {p}
                HAS THIS FIXED TITLE:
                    {t}

                DID REPLACEMENT OR JUST WARNING: {a}

            """.format(
                p=name, t=fixed_description, a="warning" if warn else "replacement"
            )
            print(msg)

        if warn:
            return None
        else:
            return fixed_description

    def fetch_external_title(
        self,
        name: str,
        steps: None | dict = None,
        verbose: bool | None = None,
    ) -> str | None:
        def _name_from_title_jsonpath(jp):
            parts = jp.split(".")
            if len(parts) > 1:
                return parts[-2]
            else:
                return jp

        if verbose is None:
            verbose = self.verbose

        if steps is None:
            steps = self.steps

        if name is not None:
            name = _name_from_title_jsonpath(name)

        external_title = steps.get(name, dict()).get("title", None)

        if external_title is not None and verbose:
            msg = """
            [ENRICHMENT] Found a title for which an external title could be fetched.
                THIS PARAMETER: {p}
                HAS THIS EXTERNAL TITLE:
                    {t}

            """.format(
                p=name,
                t=external_title,
            )
            print(msg)

        return external_title

    def fetch_flat_dict_for_field(
        self,
        name: str,
        field: str = "default",
        steps: None | dict = None,
        conservative: bool = True,
        verbose: bool | None = None,
    ) -> dict:
        if verbose is None:
            verbose = self.verbose

        value = self.mf[name]
        if (
            conservative
            and isinstance(value, (DotDict, dict))
            and len(dict(value)) == 0
        ):
            jp = name.split(".{f}".format(f=field))[0]
            flat_dict = self.collect_flat_dict(jp, field=field)
        elif not conservative:
            jp = name.split(".{f}".format(f=field))[0]
            print(jp)
            flat_dict = self.collect_flat_dict(jp, field=field)
        else:
            flat_dict = None

        if flat_dict is not None and len(flat_dict) > 0 and verbose:
            msg = """
            [ENRICHMENT] Fetched nested fields.
                THIS PARAMETER: {p}
                HAS THESE VALUES FOR {f}:
                    {d}

            """.format(
                p=name, d=flat_dict, f=field.upper()
            )
            print(msg)

        return flat_dict

    def fetch_defaults(
        self,
        name: str,
        steps: None | dict = None,
        verbose: bool | None = None,
    ) -> dict:
        return self.fetch_flat_dict_for_field(
            name=name, field="default", steps=steps, verbose=verbose, conservative=True
        )

    def fetch_available(
        self,
        name: str,
        steps: None | dict = None,
        verbose: bool | None = None,
    ) -> dict:
        return self.collect_descendants(
            name,
            key_reference_field="title",
            value_reference_field="default",
            dummy_value=None,
        )


class EnrichmentsMixin(Enrichments, FunctionDictMixin):
    def do_enrichments(self, line: str, key: str | None = None) -> str:
        """Do all enrichments. This tries to enrich the given line."""
        return self.do_functions(line, key=key, kind="enrichments")

    def add_enrichment(
        self,
        enrichment: str,
        key: str | None = None,
        **kwargs,
    ):
        return self.add_function(
            function=enrichment, key=key, kind="enrichments", **kwargs
        )



import re
from pathlib import Path

from ..parsers import Readme
from .guidelines import ChecksMixin, EnrichmentsMixin, FixesMixin
from .renderers import RendersMixin


class ReadmeImproverInspectorMixin:
    def collect_subsections(self, section: str, include_section_itself: bool = True):
        subsections = []
        inside = False
        for h, l in self.rm.headings:
            if h == section:
                # encountered the section itself (from this point on, we are 'inside the section')
                section_level = l
                inside = True
                if include_section_itself:
                    subsections.append(h)
            elif inside == True and l > section_level:
                # if we were inside, and the next section has a higher level of indent => it is an actual subsection
                subsections.append(h)
            elif inside == True and l <= section_level:
                # if we were inside, and the next section does not have a higher level of indent => it is a sibling or even parent => we are no longer 'inside the section'
                inside = False
            else:
                # we were not inside, so whatever we encounter cannot be a subsection
                pass

        return subsections

    def collect_content(
        self,
        section: str,
        include_subsections: bool = True,
        filter_headings: bool = True,
        filter_whitespace: bool = True,
    ):
        content = []

        if include_subsections:
            subsections = self.collect_subsections(section, include_section_itself=True)
        else:
            subsections = [section]

        for s in subsections:
            subsection_content = self.rm.dict().get(s, [])
            if filter_headings and len(subsection_content) > 0:
                subsection_content = subsection_content[1:]

            if filter_whitespace:
                subsection_content = [
                    l for l in subsection_content if not re.match(r"\s+", l)
                ]
            content.extend(subsection_content)

        return content

    def remove_content(
        self,
        section: str,
        include_subsections: bool = True,
        keep_headings: bool = True,
    ):
        if include_subsections:
            subsections = self.collect_subsections(section, include_section_itself=True)
        else:
            subsections = [section]

        for s in subsections:
            subsection_content = self.rm.dict().get(s, [])
            if keep_headings and len(subsection_content) > 0:
                subsection_content = subsection_content[0:1]
            else:
                subsection_content = []

            self.rm.dict()[s] = subsection_content

        return

    @property
    def platform_defined_sections(self) -> list[str]:
        return getattr(self, "_platform_defined_sections", None)

    @platform_defined_sections.setter
    def platform_defined_sections(
        self, value: list[str] | tuple[str] | set[str]
    ) -> None:
        value = list(value)

        platform_defined_sections = []
        for e in value:
            platform_defined_sections.extend(self.collect_subsections(e))

        self._platform_defined_sections = list(set(platform_defined_sections))
        return

    @property
    def user_defined_sections(self) -> list[str]:
        return list(
            set(
                [
                    s
                    for s in self.rm.dict().keys()
                    if s not in self.platform_defined_sections
                ]
            )
        )


class ReadmeImprover(
    ChecksMixin,
    FixesMixin,
    EnrichmentsMixin,
    RendersMixin,
    ReadmeImproverInspectorMixin,
):
    def __init__(
        self,
        readme_fp: str | Path | None = None,
        readme: Readme | None = None,
        platform_defined_sections=(
            "Title",
            "Inputs",
            "Outputs",
            "Pipeline overview",
            "Pipeline steps",
        ),
        verbose: bool = True,
    ):
        if readme_fp is not None:
            self.readme = Readme(readme_fp)
        elif readme is not None:
            self.readme = readme
        else:
            raise ValueError("Need a readme filepath or readme.")

        self.fp = self.rm.fp
        self.platform_defined_sections = platform_defined_sections
        self.verbose = verbose

        self.fixes = dict()
        self.checks = dict()
        self.enrichments = dict()
        self.renders = dict()
        return

    def do_improvements(
        self, input_objects: list[object], key: str | None = None
    ) -> list[object]:
        output_objects = [None] * len(input_objects)
        for idx, input_object in enumerate(input_objects):
            intermediate_object = input_object

            self.do_checks(intermediate_object, key=key)
            intermediate_object = self.do_fixes(intermediate_object, key=key)
            intermediate_object = self.do_enrichments(intermediate_object, key=key)

            output_objects[idx] = intermediate_object
        return output_objects

    def improve_lines(
        self,
        sources: list[str],
        targets: list[str] | None = None,
        key: str = "user_defined_sections",
        overwrite: bool = True,
    ):
        if targets is None:
            targets = sources

        for idx, src in enumerate(sources):
            tgt = targets[idx]

            # here is one of the few places we differ from the ManifestImprover
            old_val_src = self.rm.dict().get(src, [])  # list of lines
            old_val_tgt = self.rm.dict().get(tgt, [])  # list of lines

            new_val_src = self.do_improvements(old_val_src, key=key)

            change = old_val_src != new_val_src
            if overwrite:
                self.rm.dict()[tgt] = new_val_src
            elif change and len(old_val_tgt) == 0:
                # if there is a change and the target is empty
                self.rm.dict()[tgt] = new_val_src
            else:
                pass
        return

    def generate_section(
        self,
        target: str = "Input",
        key: str | None = None,
        erase_subsections: bool = True,
        requires_empty_target: bool = False,
    ):
        if key is None:
            key = target

        preconditions = [target in self.renders]

        if requires_empty_target:
            target_content = self.collect_content(target)
            preconditions.append(len(target_content) == 0)

        if all(preconditions):
            if erase_subsections:
                # First erase existing content before creating novel content.
                subsections = self.collect_subsections(
                    target, include_section_itself=False
                )
                for subsection in subsections:
                    self.rm.dict().pop(subsection, None)
            else:
                pass

            generated_content = self.do_renders(key=key)
            self.rm.dict()[target] = generated_content

            if self.verbose:
                msg = """
                Generated section: {s}
                """.format(
                    s=target
                )
                print(msg)

        return

    def improve_user_defined_sections(self):
        return self.improve_lines(
            sources=self.user_defined_sections, key="user_defined_sections"
        )

    def improve_platform_defined_sections(self):
        for section in self.platform_defined_sections:
            self.improve_lines(sources=[section], key=section)
        return

    def generate_platform_defined_sections(self):
        for section in self.platform_defined_sections:
            self.generate_section(target=section, erase_subsections=True)

        return

    def draft_user_defined_sections(self):
        for section in self.user_defined_sections:
            if self.verbose:
                print("Generating content for {}".format(section))
            self.generate_section(
                target=section, erase_subsections=True, requires_empty_target=True
            )

        return

    # properties
    @property
    def rm(self):
        # alias
        return self.readme

    @rm.setter
    def rm(self, value):
        # alias
        self.readme = value
        return

    @property
    def readme(self):
        return getattr(self, "_readme", None)

    @readme.setter
    def readme(self, value):
        if isinstance(value, Readme):
            self._readme = value
        else:
            raise TypeError("readme attribute only accepts Readme object.")



from pathlib import Path

# monkey patch
def _remove(self):
    from shutil import rmtree

    if self.is_file():
        self.unlink()
    else:
        rmtree(self.absolute())
    return


def _copy(self, target):
    from shutil import copy

    target = Path(target)

    assert self.is_file(), "Copy only defined on files!"

    if target.is_dir():
        target_fp = target / self.name
    else:
        target_fp = target

    copy(str(self), str(target_fp))
    return target_fp


Path.copy = _copy
Path.cp = _copy
Path.remove = _remove
Path.rm = _remove



from datetime import datetime

from batchx_dev.toolbox import Manifest, ManifestImprover, ReadmeImprover

from ..monkeypatch import Path


def make_initial_readme(readme_fp: Path, is_pipeline: bool = False):

    with open(readme_fp, "w") as f:
        f.write("")

    rmi = ReadmeImprover(readme_fp=readme_fp)

    with open(readme_fp, "w") as f:
        f.writelines(rmi.init_readme(is_pipeline=is_pipeline))
    return


def make_backup_copy(
    fp: Path, dp: Path = None, add_timestamp: bool = True, suffix: str = "backup"
):
    fp = Path(fp)

    if dp is None:
        dp = fp.parent / "backups"
    dp = Path(dp)
    dp.mkdir(exist_ok=True, parents=True)

    if add_timestamp:
        ts = datetime.now().isoformat().replace(":", "-").split(".")[0]
        suffix = "{s}-{t}".format(s=suffix, t=ts)

    bu_fp = fp.copy(dp / (fp.stem + "-" + suffix + fp.suffix))
    return bu_fp



"""
This file contains bits & pieces that need to be accessed by widely different 
parts of the codebase.

E.g., by the readme-generator & by a parser.
"""


def bx_tool_string_to_name_and_version(bx_tool_string: str) -> (str, str):
    return bx_tool_string.split(":")


from .cte import (
    INPUT_DP,
    INPUT_JSON_FP,
    MANIFEST_JSON_FP,
    OUTPUT_DP,
    OUTPUT_JSON_FP,
    ROOT,
    TMP,
)
from ..parsers import DotDict


class FileSystem(DotDict):
    def __init__(self, tool="toolname", ensure_output_dp: bool = True):
        self.fs = dict(
            root=ROOT,
            tmp=TMP,
            idp=INPUT_DP,
            odp=OUTPUT_DP,
            ifp=INPUT_JSON_FP,
            ofp=OUTPUT_JSON_FP,
            mfp=MANIFEST_JSON_FP,
        )
        self.fs["tdp"] = self.fs.get("odp") / tool
        super().__init__(self.fs, name="FileSystem")

        if ensure_output_dp:
            self.tdp.mkdir(
                parents=True, exist_ok=True
            )  # ensure tool output directory exists (not guaranteed by the batchx!)
        return



from .CTE import DEFAULT_MEMORY_MB, DEFAULT_VCPUS
from .io import load_json


# Rule configuration
def default_rule_parameters(
    name,
    config,
    fallback_timeout_s,
    fallback_vcpus=DEFAULT_VCPUS,
    fallback_memory_mb=DEFAULT_MEMORY_MB,
):
    tool_configuration = config.get("tools", dict())
    configuration = tool_configuration.get(name, dict()).copy()
    resources = configuration.pop("resources", dict())

    parameters = dict(
        name=name,
        configuration=configuration,
        timeout_s=resources.get("timeout", fallback_timeout_s),
        vcpus=resources.get("vcpus", fallback_vcpus),
        memory_mb=resources.get("memory", fallback_memory_mb),
    )
    return parameters


def get_tools(manifest_or_manifest_fp):
    def _get_name(batchx_coords):
        return batchx_coords.split("@")[-1].split(":")[0]

    def _get_canonical_name(batchx_coords):
        image_name = _get_name(batchx_coords)

        n = image_name.replace("/", "-")

        if n.startswith("bioinformatics-"):
            n = n.split("bioinformatics-")[1]

        if n.startswith("pipelines-"):
            n = n.split("pipelines-")[1]

        return n

    if isinstance(manifest_or_manifest_fp, dict):
        # you got the manifest
        manifest = manifest_or_manifest_fp
        return {
            _get_canonical_name(bx_coords): bx_coords
            for bx_coords in manifest.get("pipeline").get("steps")
        }
    else:
        # you got the manifest filepath
        manifest_fp = manifest_or_manifest_fp
        manifest = load_json(manifest_fp)
        return get_tools(manifest)



from .CTE import DAG_FP, INPUT_FP, MF_FP, OUTPUT_FP, SNAKEMAKE_EXIT_CODE_FP
from .debug import debug_print, filtered_environment
from .error import handle_unknown_exception
from .image import default_rule_parameters, get_tools
from .io import download_resource, load_json, save_json
from .run import run_bx_job, run_command



from .extract import bx_coords_and_version, get_steps_urls_dict
from .report import print_attribute


def basic_configuration_mfi(mfi, steps_dict: dict | None = None, verbose: bool = True):
    # fixes
    mfi.add_all_fixes(key="title")
    mfi.add_all_fixes(key="description")

    # checks
    mfi.add_checks(["formal_style", "avoid_the_word_image"], key="title")
    mfi.add_checks(["formal_style", "avoid_the_word_image"], key="description")

    # enrichments
    mfi.add_enrichment("document_urls", key="description")
    mfi.add_enrichment("document_urls", key="description_from_title")

    # pipeline-specific enrichments
    if mfi.mf.is_pipeline:
        if verbose:
            print("\nAdding pipeline-specific enrichments.")
        mfi.add_enrichment(
            "fill_in_fixed_description", key="fixed_description_from_name"
        )

        mfi.add_enrichment(
            "fetch_external_title", key="external_titles_from_name", steps=steps_dict
        )
    # inspect enrichments
    if verbose:
        print_attribute(mfi, "enrichments")
    return mfi


def basic_configuration_rmi(
    rmi,
    mfi,
    steps_dict: dict | None = None,
    bx_env: str = "batchx",
    include_drafts: bool = True,
    flatten=("default", "available"),
    verbose: bool = True,
):
    tool_name, tool_version = bx_coords_and_version(mfi.mf, bx_environment=bx_env)

    manifest_dict = mfi.mf.serialize(flatten=flatten).copy()
    steps_urls_dict = get_steps_urls_dict(steps_dict)

    # fixes
    rmi.add_all_fixes(key="user_defined_sections")

    # add checks
    rmi.add_checks(
        ["formal_style", "avoid_the_word_image"], key="user_defined_sections"
    )

    # consistency check needs external input
    rmi.add_check(
        "consistent_tool_reference",
        key="user_defined_sections",
        tool_name=tool_name,
        tool_version=tool_version,
    )

    # add enrichments
    rmi.add_enrichment(
        "document_urls", key="user_defined_sections"
    )  # library-defined urls (BAM, FASTA etc.)

    # manually-defined urls

    """
    extra_urls_dict = dict()
    extra_urls_dict[
        "usage tab"
    ] = "https://platform.batchx.io/test/tools/{}/usage".format(
        manifest_dict.get("name").replace("/", "%2F")
    )
    rmi.add_enrichment(
        "document_urls",
        key="user_defined_sections",
        urls_to_document=extra_urls_dict,
    )
    """

    if mfi.mf.is_pipeline:
        if verbose:
            print("\nAdding pipeline-specific enrichments to readme.")

        # pipeline-steps defined urls
        rmi.add_enrichment(
            "document_urls",
            key="user_defined_sections",
            urls_to_document=steps_urls_dict.get("docs"),
        )
        rmi.add_enrichment(
            "document_urls",
            key="Inputs",
            urls_to_document=steps_urls_dict.get("input"),
        )
        rmi.add_enrichment(
            "document_urls",
            key="Outputs",
            urls_to_document=steps_urls_dict.get("output"),
        )

    rmi.add_render("title", key="Title", manifest_dict=manifest_dict)

    if include_drafts:
        rmi.add_render("context", key="Context")
        rmi.add_render("examples", key="Examples")

    # manifest-defined sections
    if mfi.mf.is_pipeline:
        if verbose:
            print("\nAdding pipeline-specific renders to readme.")
        rmi.add_render("pipeline_inputs", key="Inputs", manifest_dict=manifest_dict)
        rmi.add_render("pipeline_outputs", key="Outputs", manifest_dict=manifest_dict)
        rmi.add_render(
            "pipeline_overview",
            key="Pipeline overview",
            steps_dict=dict(steps=steps_dict),
        )
    else:
        if verbose:
            print("\nAdding tool-specific renders to readme.")
        rmi.add_render("tool_inputs", key="Inputs", manifest_dict=manifest_dict)
        rmi.add_render("tool_outputs", key="Outputs", manifest_dict=manifest_dict)

    return rmi



from pathlib import Path
from ..utils import URLS


class ReadMeMixin:
    section_headers = dict(
        context="# Context",
        inputs="# Inputs",
        outputs="# Outputs",
        examples="# Examples",
        links="# Links",
        tool_versions="# Tool versions",
    )

    intra_section_separator = "\n\n"
    inter_section_separator = intra_section_separator + "\n"

    documentation_urls = URLS

    # Generate sections
    def readme_title(self):
        return self.title

    def readme_context(self):
        h1 = self.section_headers.get("context")
        return h1

    def readme_inputs(self):
        contents = []

        h1 = self.section_headers.get("inputs")

        if len(self.required_inputs) > 0:
            # Depending on one or multiple int
            h2 = "## Required inputs\n\nThis tool has the following **required** {i}:".format(
                i="input" if len(self.required_inputs) == 1 else "inputs"
            )
            ri = []
            for i, n in enumerate(self.required_inputs):
                l = self._single_parameter_string(name=n, index=i + 1, kind="inputs")
                ri.append(l)
            ri = self.intra_section_separator.join(ri)
            contents.extend([h2, ri])

        if len(self.optional_inputs) > 0:
            h3 = "## Optional inputs\n\nThis tool provides additional configuration through the following **optional** {i}:".format(
                i="input" if len(self.optional_inputs) == 1 else "inputs"
            )
            oi = []
            for i, n in enumerate(self.optional_inputs):
                l = self._single_parameter_string(name=n, index=i + 1, kind="inputs")
                oi.append(l)
            oi = self.intra_section_separator.join(oi)
            contents.extend([h3, oi])

        if len(contents) > 0:
            contents = [h1] + contents
            return self.intra_section_separator.join(contents)
        else:
            return ""

    def _single_parameter_string(self, name: str, index: int = 1, kind="inputs"):
        parameter = getattr(self, kind).get(name)
        def_value = parameter.get("default", None)

        l = "{i}. **`{n}`**\n\n    {d}".format(i=index, n=name, d=parameter.description)

        if def_value is not None:
            a = " (default: `{v}`).".format(v=def_value)

            # if necessary, remove full stop at end of `l`, before adding `a`.
            l = l.rstrip(".") + a
        else:
            pass

        return l

    def readme_outputs(self):
        contents = []

        h1 = self.section_headers.get("outputs")

        if len(self.required_outputs) > 0:
            h2 = "## Required outputs\n\nThis tool will **always** provide the following {o}:".format(
                o="output" if len(self.required_outputs) == 1 else "outputs"
            )
            ro = []
            for i, n in enumerate(self.required_outputs):
                l = self._single_parameter_string(name=n, index=i + 1, kind="outputs")
                ro.append(l)
            ro = self.intra_section_separator.join(ro)
            contents.extend([h2, ro])

        if len(self.optional_outputs) > 0:
            h3 = "## Optional outputs\n\nThis tool will **optionally** provide the following {o}:".format(
                o="output" if len(self.optional_outputs) == 1 else "outputs"
            )
            oo = []
            for i, n in enumerate(self.optional_outputs):
                l = self._single_parameter_string(name=n, index=i + 1, kind="outputs")
                oo.append(l)
            oo = self.intra_section_separator.join(oo)
            contents.extend([h3, oo])

        if len(contents) > 0:
            contents = [h1] + contents
            return self.intra_section_separator.join(contents)
        else:
            return ""

    def readme_examples(self):
        h1 = self.section_headers.get("examples")
        return h1

    def readme_links(self):
        h1 = self.section_headers.get("links")
        return h1

    def readme_versions(self):
        h1 = self.section_headers.get("tool_versions")
        return h1

    # Write Readme
    def write_sections(self):
        generators = dict(
            title=self.readme_title,
            context=self.readme_context,
            inputs=self.readme_inputs,
            outputs=self.readme_outputs,
            examples=self.readme_examples,
            links=self.readme_links,
            tool_versions=self.readme_versions,
        )
        sections = {k: g() for k, g in generators.items()}
        return sections

    def write(self, sections, file: str | Path = "readme.md"):

        if file is None:
            return self.inter_section_separator.join(list(sections.values()))
        else:
            with open(file, "w") as f:
                print(self.write(sections, file=None), file=f)
            return "Wrote auto-generated readme stub to {}".format(file)

    # Read Readme
    def read_sections(self, file: str | Path = "readme.md"):
        with open(file, "r") as f:
            lines = f.readlines()

        sections = {k: "" for k in self.section_headers.keys()}
        sections["title"] = ""
        active_section = "title"

        for l in lines:
            if l.startswith("#"):
                for k, v in self.section_headers.items():
                    if l.startswith(v):
                        active_section = k

            sections[active_section] = sections[active_section] + l

        sections = {k: s.rstrip("\n") for k, s in sections.items()}

        return sections

    # alias
    def readme(self, file: str | Path = "readme.md", overwrite=False):
        if file is None:
            sections = self.write_sections()
            return self.write(sections, file=file)
        elif overwrite or not Path(file).exists():
            sections = self.write_sections()
            return self.write(sections, file=file)
        else:
            old_sections = self.read_sections(file)
            new_sections = self.write_sections()

            sections = dict(
                title=new_sections.get("title"),
                context=old_sections.get("context"),
                inputs=new_sections.get("inputs"),
                outputs=new_sections.get("outputs"),
                examples=old_sections.get("examples"),
                links=old_sections.get("links"),
                tool_versions=old_sections.get("tool_versions"),
            )
            return self.write(sections, file=file)



def print_attribute(obj, attr: str = "available_checks"):
    result = getattr(obj, attr, dict())
    print("{}\n".format(attr.replace("_", " ").capitalize()))

    for k in result:
        print("    - {}".format(k))
    return


def print_available_improvements(
    obj,
    attributes=(
        "available_fixes",
        "available_checks",
        "available_enrichments",
        "available_renders",
    ),
    level=2,
):
    header = "Available improvements {}".format(obj.__class__.__name__)
    print_report_header(header, level=level)
    for a in [
        "available_fixes",
        "available_checks",
        "available_enrichments",
        "available_renders",
    ]:
        print_attribute(obj, a)
    return


def print_report_header(header: str, level: int = 1):
    spacing = max(min(3, 3 - level), 1)
    prefix = ("\n" * spacing) + ("#" * level)
    suffix = "\n" * spacing

    print(" ".join([prefix, header, suffix]))
    return



import json
import os
import subprocess as sp
import sys
import traceback
from functools import partial
from pathlib import Path
from time import sleep

from batchx import bx

from .CTE import DEFAULT_MEMORY_MB, DEFAULT_VCPUS
from .debug import debug_print
from .error import (
    handle_called_process_error,
    handle_unknown_exception,
    handle_value_error,
)
from .io import save_json


def handle_custom_exception(
    stderr: str,
    custom_exceptions: list | None = None,
    exit_on_custom_exception: bool = True,
):
    if custom_exceptions is None:
        custom_exceptions = []

    for custom_exception in custom_exceptions:
        if custom_exception.is_applicable(stderr):
            debug_print(
                "{id} Stderr".format(id=custom_exception.IDENTIFIER), stderr, long=False
            )
            if exit_on_custom_exception:
                sys.exit(custom_exception.BX_ERROR_CODE)
            else:
                pass
        else:
            pass
    return


def check_for_acceptable_exit_code(
    identifier: str = "",
    error_message: str | None = None,
    exit_code: int | None = None,
    acceptable_exit_codes: tuple[int, ...] = (0,),
    verbose: bool = True,
    raise_error: bool = False,
):
    debug_print(
        "{} | Exit code".format(identifier),
        exit_code,
        verbose=verbose,
        long=False,
    )

    if exit_code not in acceptable_exit_codes:
        msg = """
        Exit code {c} not in acceptable exit codes {ac}.
        StdErr: {error_msg}
        """.format(
            error_msg=error_message,
            c=exit_code,
            ac=acceptable_exit_codes,
        )
        if raise_error:
            raise ValueError(msg) from None
        else:
            debug_print(
                "Inacceptable exit code error message",
                msg,
                verbose=verbose,
            )
            return exit_code
    else:
        return exit_code


def get_bx_submit_request(
    job_service,
    identifier: str,
    tool: str,
    inputs: dict,
    environment: str | None = None,
    vcpus=DEFAULT_VCPUS,
    memory_mb=DEFAULT_MEMORY_MB,
    timeout_s: int = 15 * 60,
    verbose: bool = True,
):

    request = job_service.SubmitRequest(
        environment=environment,
        image=tool,
        vcpus=vcpus,
        memory=memory_mb,
        timeout=timeout_s,
        input_json=json.dumps(inputs),
    )
    debug_print("{} Request".format(identifier), request, verbose=verbose)
    return request


def inspect_bx_response(
    error_message: str | None,
    exit_code: int | None,
    identifier: str,
    acceptable_exit_codes: tuple[int, ...] = (0,),
    inacceptable_exit_code_raise_error: bool = False,
    verbose: bool = True,
):
    if error_message not in {None, ""}:
        debug_print(
            "{} Error message".format(identifier),
            error_message,
            verbose=verbose,
            long=False,
        )

    # check exit code
    exit_code = check_for_acceptable_exit_code(
        identifier=identifier,
        error_message=error_message,
        exit_code=exit_code,
        acceptable_exit_codes=acceptable_exit_codes,
        raise_error=inacceptable_exit_code_raise_error,
    )
    return exit_code


def run_command(
    command: str,
    identifier: str = "",
    output_fp: Path | str | None = None,
    verbose: bool = True,
    cwd=None,
    shell: bool = True,
    custom_exceptions: list | None = None,
    acceptable_exit_codes: tuple[int, ...] = (0,),
    raise_unknown_exception: bool = True,
    raise_custom_exception: bool = True,
    raise_exitcode_exception: bool = True,
):

    debug_print("{} command".format(identifier), command, verbose=verbose, long=False)
    run_subprocess = partial(
        sp.run,
        command,
        cwd=cwd,
        shell=shell,
        stderr=sp.PIPE,
        encoding="utf-8",
    )

    try:
        if output_fp is None:
            completed_subprocess = run_subprocess()
        else:
            with open(output_fp, "w") as f:
                completed_subprocess = run_subprocess(stdout=f)
    except Exception as e:
        handle_unknown_exception(
            exception=e, identifier=identifier, reraise=raise_unknown_exception
        )

    # see if the output is good or we need to trigger an error of some kind.
    if verbose:
        debug_print("stdout", completed_subprocess.stdout)
        debug_print("stderr", completed_subprocess.stderr)

    handle_custom_exception(
        completed_subprocess.stderr,
        custom_exceptions=custom_exceptions,
        exit_on_custom_exception=raise_custom_exception,
    )

    exit_code = check_for_acceptable_exit_code(
        identifier=identifier,
        error_message=completed_subprocess.stderr,
        exit_code=completed_subprocess.returncode,
        acceptable_exit_codes=acceptable_exit_codes,
        raise_error=raise_exitcode_exception,
    )
    return exit_code


def run_bx_job(
    tool: str,
    inputs: dict,
    output_fp=None,
    environment: str | None = None,
    vcpus=DEFAULT_VCPUS,
    memory_mb=DEFAULT_MEMORY_MB,
    timeout_s: int = 15 * 60,
    verbose: bool = True,
    reraise: bool = False,
    acceptable_exit_codes: tuple[int, ...] = (0,),
    inacceptable_exit_code_raise_error: bool = False,
):
    # Initialize
    output_json = dict()
    if environment is None:
        environment = os.environ["BATCHX_ENV"]

    identifier = "{t} bx job submit".format(t=tool)

    bx.connect()
    job_service = bx.JobService()

    submit_request = get_bx_submit_request(
        job_service,
        identifier=identifier,
        tool=tool,
        inputs=inputs,
        environment=environment,
        vcpus=vcpus,
        memory_mb=memory_mb,
        timeout_s=timeout_s,
        verbose=verbose,
    )

    try:
        run_response = job_service.Run(submit_request)
        debug_print("{} Response".format(identifier), run_response, verbose=verbose)

        # check potential error message
        error_message = getattr(run_response, "error_message", None)
        exit_code = getattr(run_response, "exit_code", 1)
        output_json = json.loads(getattr(run_response, "output_json", "{}"))
    except Exception as e:
        handle_unknown_exception(exception=e, identifier=identifier, reraise=reraise)

    if output_fp is not None:
        save_json(output_json, output_fp)
    else:
        pass

    exit_code = inspect_bx_response(
        error_message=error_message,
        exit_code=exit_code,
        identifier=identifier,
        acceptable_exit_codes=acceptable_exit_codes,
        inacceptable_exit_code_raise_error=inacceptable_exit_code_raise_error,
        verbose=verbose,
    )

    return exit_code, error_message, output_json



from ..parsers import Input, Manifest


class CommandKindMixin:

    # writer functions
    def get_cmd(self, input: Input, sort=True) -> list[str]:
        """
        Generate command array.

        Args:
            input (Input): Input object (encodes an input json)
            sort (bool, optional): Whether or not to sort the element of the command alphabetically.
                                    If set to True, the order in which the commands were added to the CommandBuilder is
                                    maintained when generating the final command. Defaults to True.

        Returns:
            list[str]: Array of command parts.
        """

        input = self.check_constraints(input, self.manifest)
        input = self.do_actions(input, self.manifest)

        tool_cmd = [s for s in str(self.tool).split() if len(s) > 0]

        kval_cmds, kval_keys = self._get_keyvalue_cmd(input)
        flag_cmds, flag_keys = self._get_flag_cmd(input)
        args_cmds, args_keys = self._get_argument_cmd(input)
        cmds = kval_cmds + flag_cmds + args_cmds

        if sort:
            keys = kval_keys + flag_keys + args_keys
            d = {k: c for k, c in zip(keys, cmds)}
            cmds = [d.get(k, None) for k in self.commands]
            cmds = [c for c in cmds if c is not None]

        cmds = [e for c in cmds for e in c if len(e) > 0]

        return tool_cmd + cmds

    def _get_keyvalue_cmd(self, input: Input, force=False):
        cmds = []
        keys = []
        for key in self.kinds["keyval"]:
            v = getattr(input, key, None)
            if v is not None or force:

                if isinstance(v, list):
                    part = [self.commands[key], *v]
                else:
                    part = [self.commands[key], v]

                part = ["{e}".format(e=e) for e in part]
                part = [e for e in part if len(e) > 0]

                cmds.append(part)
                keys.append(key)
        return cmds, keys

    def _get_flag_cmd(self, input: Input, force=False):
        cmds = []
        keys = []
        for key in self.kinds["flag"]:
            if getattr(input, key, False) or force:
                part = [self.commands[key]]

                part = ["{e}".format(e=e) for e in part]
                part = [e for e in part if len(e) > 0]
                cmds.append(part)
                keys.append(key)
        return cmds, keys

    def _get_argument_cmd(self, input: Input, force=False):
        cmds = []
        keys = []
        for key in self.kinds["argument"]:
            if getattr(input, key, None) is not None or force:
                v = getattr(input, key, None)

                if isinstance(v, list):
                    part = [*v]
                else:
                    part = [v]

                part = ["{e}".format(e=e) for e in part]
                part = [e for e in part if len(e) > 0]
                cmds.append(part)
                keys.append(key)
        return cmds, keys

    # assignment
    def set_kind(self, kind: str = "keyval", key: str | None = None):
        if key is None:
            key = self.active_key

        # check if this key was already assigned to a kind
        for keys_from_a_specific_kind in self.kinds.values():
            if key in keys_from_a_specific_kind:
                keys_from_a_specific_kind.remove(key)

        # assign this key to a kind
        if self.kinds.get(kind, None) is not None:
            self.kinds[kind].add(key)
        else:
            self.kinds[kind] = set([key])

        self.active_key = key
        return self

    # alias
    def add_kind(self, kind: str = "keyval", key: str | None = None):
        # alias
        return self.set_kind(kind=kind, key=key)



from pathlib import Path

from .dotdict import DotDict
from .json_mixin import JSONLoaderMixin


class Input(DotDict, JSONLoaderMixin):
    def __init__(
        self,
        input_fp: str | Path | None = None,
        input_dict: dict | None = None,
        jsonpaths: bool = False,
    ):
        if input_fp is not None:
            self.fp = input_fp
            self.load()  # ensures existence of self.json attribute
        elif input_dict is not None:
            self.load_dict(input_dict)  # ensures existence of self.json attribute
        else:
            raise ValueError("Need input filepath or input dictionary.")

        super().__init__(self.json, name="Input", jsonpaths=jsonpaths)
        return



class AutoBuiltIns:
    @staticmethod
    def _head_tail_json_path(json_path):
        components = json_path.split(".")

        head = components[0]
        tail = ".".join(components[1:]) if len(components) > 1 else None
        return head, tail

    # dict built-ins
    def __getitem__(self, __name__):
        # 1: get the item
        # 2: split path
        # 3: skip properties and get item

        def _jsonpaths_applicable():
            return self.__jsonpaths__ and "properties" in self.keys()

        def _action_01(name):
            return self.dict()[name]

        def _action_02(name):

            head, tail = self._head_tail_json_path(name)

            if tail is not None:
                return self.dict()[head][tail]
            else:
                raise KeyError("KeyError: '{n}'".format(n=head))

        def _action_03(name):
            return self.dict()["properties"][name]

        if self.__jsonpaths__:
            actions = [_action_01, _action_02, _action_03]
        else:
            actions = [_action_01]

        value = None
        no_error = False
        error = None
        for counter, action in enumerate(actions):
            try:
                value = action(__name__)
                no_error = True
                break
            except KeyError as e:
                error = e
                # dotdicts do not throw keyerrors, only the underlying dotdict.dict() do!
                continue

        if no_error:
            return value
        else:
            raise KeyError(error)

    def __setitem__(self, __name__: str, __value__):

        if self.__jsonpaths__:
            head, tail = self._head_tail_json_path(__name__)
            if tail is not None:
                __name__ = head

                __present_value__ = self.get(head, dict())
                __updated_value__ = {tail: __value__}

                assert hasattr(
                    __present_value__, "items"
                ), "Your jsonpath is leading to the expansion of an object that is not a mapping: is the path correct?"
                __value__ = {**__present_value__, **__updated_value__}
        else:
            pass

        if self.__recursive__ and isinstance(__value__, dict):
            __value__ = DotDict(
                __value__,
                name=__name__,
                recursive=self.__recursive__,
                jsonpaths=self.__jsonpaths__,
            )

        return self.dict().__setitem__(__name__, __value__)

    # attribute built-ins
    def __getattr__(self, __name__: str, default=None):
        try:
            return self.__getitem__(__name__)
        except KeyError:
            return default

    def __setattr__(self, __name__: str, __value__):
        if self.__overrride_setattr__:
            self.__setitem__(__name__, __value__)
        else:
            super().__setattr__(__name__, __value__)
            return


class DotDict(AutoBuiltIns):
    __overrride_setattr__ = False

    def __init__(
        self,
        dictionary,
        name: str = "dotdict",
        recursive: bool = True,
        jsonpaths: bool = False,
    ):
        self.__name__ = name
        self.__recursive__ = recursive
        self.__jsonpaths__ = jsonpaths
        self.__dictionary__ = dict()
        for k, v in dictionary.items():
            self.__setitem__(k, v)

        self.__overrride_setattr__ = True
        return

    # representations
    def __repr__(self):
        return "{} {}".format(self.__name__, list(self.keys()))

    # serialize (to dump to json)
    def serialize(self, flatten=("")):
        def _serialize(o):
            d = dict()
            for k, v in o.items():
                if isinstance(v, dict):
                    v = _serialize(v)
                elif isinstance(v, DotDict) and k in flatten:
                    v = v.flatten()
                elif isinstance(v, DotDict) and k not in flatten:
                    v = _serialize(v)
                else:
                    pass
                d[k] = v
            return d

        return _serialize(self)

    # mirror dict behavior
    def dict(self):
        return self.__dictionary__

    # dict-like behavior
    def keys(self):
        return self.dict().keys()

    def values(self):
        return self.dict().values()

    def items(self):
        return self.dict().items()

    def get(self, key, default=None):
        return self.dict().get(key, default)

    def pop(self, key, default=None):
        return self.dict().pop(key, default)

    # jsonpaths
    def collect_jsonpaths(self, prefix: str = ""):
        def _collect_jsonpaths(dd, prefix=""):
            keys = []
            prefix = "{p}.".format(p=prefix) if len(prefix) > 0 else ""
            for k, v in dd.items():
                key = "{p}{k}".format(p=prefix, k=k)

                if hasattr(v, "items") and len(v.items()) > 0:
                    keys.extend(_collect_jsonpaths(v, prefix=key))
                else:
                    keys.append(key)
            return keys

        return _collect_jsonpaths(self, prefix="")

    def flatten(self, prefix: str = ""):
        return {k: self[k] for k in self.collect_jsonpaths(prefix=prefix)}

    def jsonpaths(self):
        # alternative for keys() for a complete iteration
        return self.collect_jsonpaths()

    @property
    def jsonpaths_enabled(self):
        return self.__jsonpaths__



"""
Functions to preprocess/extract useful information for MFI and RMI
"""


def get_steps_urls_dict(
    steps_dict: dict, categories: tuple[str] = ("input", "output", "docs")
):
    steps_urls_dict = dict(input=dict(), output=dict(), docs=dict())
    for k, v in steps_dict.items():
        steps_urls_dict["docs"][k.capitalize()] = v.get("urls").get("docs")

        key = "**`{k}`**".format(k=k)
        for category in categories:
            steps_urls_dict[category][key] = v.get("urls").get(category)
    return steps_urls_dict


def bx_coords_and_version(mf, bx_environment: str = "batchx"):
    bx_coords = "{env}@{n}".format(env=bx_environment, n=mf.name)
    return bx_coords, mf.version



from pathlib import Path


from .dotdict import DotDict
from .json_mixin import JSONLoaderMixin
from .readme_mixin import ReadMeMixin


class Manifest(DotDict, JSONLoaderMixin, ReadMeMixin):
    def __init__(self, manifest_fp: str | Path, jsonpaths: bool = False):
        self.fp = manifest_fp
        self.load()  # ensures existence of self.json attribute
        super().__init__(self.json, name="Manifest", jsonpaths=jsonpaths)
        return

    # PARAMETERS
    @property
    def parameters(self):
        return {**self.inputs, **self.outputs}

    # INPUTS
    @property
    def inputs(self):
        return self.schema.input.properties  # shortcut

    @property
    def required_inputs(self) -> list[str]:
        return [
            i
            for i in self.input_names
            if (
                (getattr(self.inputs, i).required)
                or (i in getattr(self.schema.input, "required", []))
            )
        ]

    @property
    def optional_inputs(self) -> list[str]:
        return [i for i in self.input_names if i not in self.required_inputs]

    @property
    def input_names(self):
        return self.inputs.keys()

    @property
    def default_inputs(self):
        return {k: v.get("default", None) for k, v in self.inputs.items()}

    # OUTPUTS
    @property
    def outputs(self):
        return self.schema.output.properties  # shortcut

    @property
    def required_outputs(self):
        return [
            o
            for o in self.output_names
            if (
                (getattr(self.outputs, o).required)
                or (o in getattr(self.schema.output, "required", []))
            )
        ]

    @property
    def optional_outputs(self):
        return [o for o in self.output_names if o not in self.required_outputs]

    @property
    def output_names(self):
        return self.outputs.keys()

    @property
    def default_outputs(self):
        return {k: v.get("default", None) for k, v in self.outputs.items()}

    # PIPELINE
    @property
    def is_pipeline(self):
        return self.pipeline is not None

    @property
    def sample(self):
        if self.is_pipeline:
            return self.schema.input.properties.sample.properties  # shortcut
        else:
            return None

    @property
    def tools(self):
        if self.is_pipeline:
            return self.schema.input.properties.tools.properties  # shortcut
        else:
            return None



import json
from pathlib import Path


class JSONLoaderMixin:
    def load_dict(self, dictionary: dict):
        # put python dict through json encoder and decoder
        contents = json.dumps(dictionary)
        self._json = json.loads(contents)
        return

    def load(self):
        with open(self.fp, "r") as f:
            contents = f.read()

        self._json = json.loads(contents)
        return

    def save(self, fp: str | Path | None = None, return_fp: bool = False):
        if fp is None:
            fp = self.fp

        with open(fp, "w") as f:
            json.dump(self.serialize(), f, indent=4)

        print("Saved file to: {}".format(fp))
        if return_fp:
            return fp
        else:
            return

    # properties
    @property
    def fp(self):
        return getattr(self, "_fp", None)

    @fp.setter
    def fp(self, value: str | Path):
        assert isinstance(value, str | Path), "Must be filepath."
        self._fp = Path(value)
        return

    @property
    def json(self):
        return getattr(self, "_json", None)

    @json.setter
    def json(self, value: str | Path):
        self.fp = value
        self.load()
        return



from functools import partial
from typing import Callable


class Constraints:
    @property
    def preconfigured_constraints(self):
        return dict(requires=self.requires)

    def requires(self, input, manifest, value, k=None, v=None):

        assert k in manifest.input_names, "Unknown Key {k}: not in manifest.".format(
            k=k
        )
        ip_val = getattr(input, k, None)
        mf_val = manifest.default_inputs.get(k, None)

        if v is None:
            r = k in input.keys()
            msg = """
            Checking library constraint "requires"
                Is key `{k}` explicitly given in input?
                (value unimportant)

            Result: {r}
            """.format(
                k=k, r=r
            )
            print(msg)
            return r
        elif k in input.keys():
            r = ip_val == v
            msg = """
            Checking library constraint "requires".
            Key {k} is explicitly set in input.
                input[{k}] == {v}?

            Result: {r}
            """.format(
                k=k, r=r, v=v
            )
            print(msg)
            return r
        else:
            r = mf_val == v
            msg = """
            Checking library constraint "requires".
            Key {k} is not explicitly set in input (check manifest-specified default)
                manifest[{k}]["default"] == {v}?

            Result: {r}                
            """.format(
                k=k, v=v, r=r
            )
            print(msg)
            return r


class ConstraintMixin(Constraints):
    def check_constraints(self, input, manifest):
        """Check all constraints."""

        all_keys = input.jsonpaths() if input.jsonpaths_enabled else input.keys()
        for key in all_keys:
            self.check_constraints_for_one_key(input, manifest, key)
        return input

    def check_constraints_for_one_key(self, input, manifest, key: str):
        relevant_constraints = self.constraints.get(key, [])
        for constraint in relevant_constraints:
            val = getattr(input, key)  # really needs to be *inside* the loop
            if self.verbose:
                msg = """
                CHECK CONSTRAINT
                ----------------
                Constraint: {c}
                for Key:    {k}
                with Value: {v}
                """.format(
                    k=key, v=val, c=constraint
                )
                print(msg)
            if constraint(input, manifest, val) is True:
                pass
            else:
                msg = """
                Constraint {c} 
                related to key {k} not satisfied in this input.
                """.format(
                    c=constraint, k=key
                )
                raise ValueError(msg)
        return True

    def add_constraint(
        self, constraint: Callable | str = None, key: str | None = None, **kwargs
    ):
        if key is None:
            key = self.active_key

        # obtain constraint-function
        if isinstance(constraint, str):
            if constraint in self.preconfigured_constraints:
                constraint = self.preconfigured_constraints.get(constraint)
            else:
                raise ValueError("Did not recognize action {c}".format(c=constraint))
        elif isinstance(constraint, Callable):
            constraint = constraint
        else:
            raise ValueError("Constraint needs to be string or callable.")

        # partially initialize function if desired
        if len(kwargs) > 0:
            constraint = partial(constraint, **kwargs)

        # record function in constraints dict
        if self.constraints.get(key, None) is not None:
            self.constraints[key].append(constraint)
        else:
            self.constraints[key] = [constraint]

        self.active_key = key
        return self



import os
import pprint


def debug_print(
    identifier: str,
    value: object | None = None,
    verbose=True,
    flush=True,
    long: bool = True,
):
    if verbose:
        if value is None:
            msg = "{i}".format(i=identifier)
        else:
            if isinstance(value, dict):
                value = "\n{}".format(pprint.pformat(value, indent=2))

            if long:
                msg = "Variable name:  {i}\nVariable value: {v}".format(
                    i=identifier, v=value
                )
            else:
                msg = "{i}: {v}".format(i=identifier, v=value)

        print(msg, flush=flush)
    return


def filtered_environment(
    add_toolbox_version: bool = True,
    prefixes: tuple[str] | set[str] | list[str] = ("BX", "PYTHON", "PATH", "PWD"),
):
    # NB do not add BATCHX to prefixes, contains private info.
    d = {k: v for k, v in os.environ.items() if any([e in k for e in prefixes])}

    if add_toolbox_version:
        try:
            from batchx_dev import __version__ as BX_PYTHON_TOOLBOX_VERSION

            d["BX_PYTHON_TOOLBOX_VERSION"] = BX_PYTHON_TOOLBOX_VERSION
        except Exception as e:
            debug_print(
                "Fetch BX_PYTHON_TOOLBOX_VERSION | Unexpected Exception.",
                e,
                verbose=True,
            )
    return d



from functools import partial

from ..functiondict_mixin import FunctionDictMixin


class Checks:
    @property
    def available_checks(self):
        return dict(
            formal_style=self.formal_style,
            avoid_the_word_image=self.avoid_the_word_image,
            consistent_tool_reference=self.consistent_tool_reference,
        )

    def formal_style(self, line: str, verbose: bool | None = None) -> str:
        if verbose is None:
            verbose = self.verbose

        if line.find(" you ") > -1 or line.find(" we ") > -1:
            if verbose:
                msg = """
                [STYLE GUIDE] Prefer formal style: avoid "you" or "we".
                    CONSIDER REWRITING THIS: 
                        {l}
                """.format(
                    l=line
                )
                print(msg)
        else:
            pass
        return line

    def avoid_the_word_image(self, line: str, verbose: bool | None = None) -> str:
        if verbose is None:
            verbose = self.verbose

        if line.find(" image") > -1:
            if verbose:
                msg = """
                [STYLE GUIDE] Avoid using the word "image" to refer to a tool.
                    DOUBLE-CHECK THIS: 
                        {l}
                """.format(
                    l=line
                )
                print(msg)
        else:
            pass
        return line

    def consistent_tool_reference(
        self,
        line: str,
        tool_name: str,
        tool_version: str,
        window: int = 40,
        verbose: bool | None = None,
    ) -> str:
        if verbose is None:
            verbose = self.verbose

        def _fetch_excerpt(line: str, idx: int, window: int = 40):
            start_idx = max(idx - window, 0)
            stop_idx = min(idx + window, len(line) - 1)
            excerpt = line[start_idx:stop_idx]
            return excerpt

        tool_name_str = "{n}:".format(n=tool_name)
        tool_name_and_version_str = "{tns}{v}".format(tns=tool_name_str, v=tool_version)
        if (
            line.find(tool_name_str) > -1
            and not line.find(tool_name_and_version_str) > -1
        ):
            idx = line.find(tool_name_str)
            excerpt = _fetch_excerpt(line, idx, window=window)

            if verbose:
                msg = """
                [INCONSISTENCY ALERT] Version number in manifest does not match a detected tool reference
                    COMPARE THIS TEXT: {e}

                    MANIFEST VERSION: {v}
                """.format(
                    e=excerpt, v=tool_version
                )
                print(msg)
        else:
            pass
        return line


class ChecksMixin(Checks, FunctionDictMixin):
    def do_checks(self, line: str, key: str | None = None) -> None:
        """Do all checks. This will not transform anything, but warns you if something maybe could be done better."""

        return self.do_functions(line, key=key, kind="checks")

    def add_check(
        self,
        check: str,
        key: str | None = None,
        **kwargs,
    ):
        return self.add_function(function=check, key=key, kind="checks", **kwargs)

    def add_checks(
        self,
        checks: list[str],
        key: str | None = None,
        **kwargs,
    ):
        for check in checks:
            self.add_check(check, key=key, **kwargs)
        return

    def add_all_checks(self, key: str | None = None, **kwargs):
        """
        Associate all available checks with a particular key.

        Args:
            key (str | None, optional): Key for this collection of checks. Defaults to None.
        """
        return self.add_checks(self.available_checks, key=key, **kwargs)



import json
import shutil
import urllib.request as request
from contextlib import closing
from pathlib import Path

import requests
from batchx import bx

from .debug import debug_print


# load/save
def load_json(fp: Path | str):
    with open(fp, "r") as f:
        contents = f.read()
    return json.loads(contents)


def save_json(
    contents,
    fp: Path | str,
    verbose: bool = True,
):
    fp = Path(fp)

    dp = fp.parent
    dp.mkdir(parents=True, exist_ok=True)

    if isinstance(contents, str):
        contents = json.dumps(contents, default=str)

    with open(fp, "w+") as f:
        json.dump(contents, f, default=str)

    debug_print("Saved json to {}".format(fp), verbose=verbose)
    return


# download protocols
def download_resource(url: Path | str, fp: Path | str, verbose: bool = True, **kwargs):
    actions = dict(
        http=download_http,
        https=download_http,
        s3=download_s3,
        ftp=download_ftp,
        bx=download_bx,
    )

    protocol = Path(url).parts[0].split(":")[0]
    return actions.get(protocol)(url=url, fp=fp, verbose=verbose, **kwargs)


def download_encode(url: str, fp: Path | str):
    """Download file from the ENCODE portal."""

    def _obtain_json(url: str):
        headers = dict(
            accept="application/json"
        )  # Force return from the server in JSON format
        response = requests.get(url, headers=headers)  # GET the object
        return response.json()

    encode_json = _obtain_json(url)
    return download_s3(url=encode_json.get("s3_uri"), fp=fp)


def download_s3(url: Path | str, fp: Path | str, verbose: bool = True):
    try:
        import boto3
    except:
        raise ImportError("You should install boto3 if you want this functionality.")

    def _bucket_key_from_s3_uri(uri: Path | str):
        uri = str(uri)
        return uri.split("/", 2)[-1].split("/", 1)

    url = str(url)

    s3 = boto3.client("s3")
    bucket, key = _bucket_key_from_s3_uri(url)
    s3.download_file(bucket, key, str(fp))
    return


def download_bx(url: Path | str, fp: Path | str, env="test", verbose: bool = True):
    url = str(url)

    bx.connect()
    fs = bx.FilesystemService()
    request = fs.DownloadPresignedRequest(environment=env, path=url)
    response = fs.DownloadPresigned(request)

    return download_http(url=response.url, fp=fp, verbose=verbose)


def download_ftp(url: Path | str, fp: Path | str, verbose: bool = True):
    url = str(url)
    with closing(request.urlopen(url)) as src:
        with open(fp, "wb") as dst:
            shutil.copyfileobj(src, dst)
    return


def download_http(url: Path | str, fp: Path | str, verbose: bool = True):
    url = str(url)

    with open(fp, "wb") as f:
        f.write(requests.get(url).content)

    if verbose:
        fp = Path(fp)
        msg = """
        curl {url} --output {fn}
        """.format(
            url=url, fn=fp.name
        )
        print(msg)
    return



import json
from batchx import bx as bx_api


class BxImage:
    def __init__(self, image_coords: str):
        bx_api.connect()
        self.image_coords = image_coords
        self.load()
        return

    def load(self):
        image_service = bx_api.ImageService()
        req = image_service.GetImageRequest(image=self.image_coords)

        self.GetImageResponse = image_service.GetImage(req)
        return

    def _fields_to_dict(self, pb):
        if hasattr(pb, "ListFields"):
            return {k.name: self._fields_to_dict(f) for (k, f) in pb.ListFields()}
        else:
            return pb

    @property
    def response_dict(self):
        if not hasattr(self, "_response_dict"):
            self._response_dict = self._fields_to_dict(self.GetImageResponse)
        return self._response_dict

    @property
    def image_dict(self):
        return self.response_dict.get("image")

    @property
    def manifest_dict(self):
        if not hasattr(self, "_manifest_dict"):
            self._manifest_dict = json.loads(self.image_dict.get("manifest"))
        return self._manifest_dict

    @property
    def environment(self):
        return self.image_dict.get("environment")

    @property
    def name(self):
        return self.image_dict.get("name")

    @property
    def canonical_name(self):
        n = self.name.replace("/", "-")

        if n.startswith("bioinformatics-"):
            n = n.split("bioinformatics-")[1]

        if n.startswith("pipelines-"):
            n = n.split("pipelines-")[1]

        return n

    @property
    def version(self):
        return self.image_dict.get("version")

    @property
    def description(self):
        return self.image_dict.get("description")

    @property
    def title(self):
        return self.description

    @property
    def url(self):
        return "https://platform.batchx.io/{e}/tools/{n}/{v}".format(
            e=self.environment, n=self.name.replace("/", "%2F"), v=self.version
        )

    @property
    def urls(self):
        return dict(
            base=self.url,
            docs="{u}/documentation".format(u=self.url),
            input="{u}/input".format(u=self.url),
            output="{u}/output".format(u=self.url),
        )



import json
import re
from pathlib import Path

import mistletoe
from mistletoe.ast_renderer import ASTRenderer


class MarkdownLoaderMixin:
    _INITIAL_SECTION_NAME = "Title"

    # load-methods
    def load(self):
        self.load_lines()
        self.load_ast()
        self.load_dict()
        return

    def load_lines(self):
        with open(self.fp, "r") as f:
            lines = f.readlines()

        self._lines = lines
        return

    def load_ast(self):

        with open(self.fp, "r") as f:
            ast = mistletoe.markdown(f, renderer=ASTRenderer)

        self._ast = json.loads(ast)

        return

    def load_dict(self):
        contents = dict()
        active_section = self._INITIAL_SECTION_NAME
        for l in self._lines:
            for (h, level), r in self.regexes.items():
                if re.match(r, l):
                    active_section = h
                else:
                    pass

            if contents.get(active_section) is None:
                contents[active_section] = [l]
            else:
                contents[active_section].append(l)

        self.__dictionary__ = contents
        return

    # save-methods
    def serialize(self):
        lines = []
        for k, l in self.items():
            lines.extend(l)
        return lines

    def save(self, fp: str | Path | None = None, return_fp: bool = False):
        if fp is None:
            fp = self.fp

        with open(fp, "w") as f:
            f.writelines(self.serialize())

        print("Saved file to: {}".format(fp))
        if return_fp: 
            return fp
        else:
            return 

    # mirror dict behavior
    def dict(self):
        return self.__dictionary__

    # dict-like behavior
    def keys(self):
        return self.dict().keys()

    def values(self):
        return self.dict().values()

    def items(self):
        return self.dict().items()

    def get(self, key, default=None):
        return self.dict().get(key, default)

    # properties
    @property
    def fp(self):
        return getattr(self, "_fp", None)

    @fp.setter
    def fp(self, value: str | Path):
        assert isinstance(value, str | Path), "Must be filepath."
        self._fp = Path(value)
        return

    @property
    def headings(self):
        def _get_headings(children: list[dict] | dict):
            if isinstance(children, dict):
                children = [children]
            headings = []
            for c in children:
                c_type = c.get("type")
                c_level = c.get("level")
                c_children = c.get("children", [])
                if c_type == "Heading":
                    header = c_children[0].get("content")
                    headings.append((header, c_level))
                else:
                    pass

                headings.extend(_get_headings(c_children))
            return headings

        headings = _get_headings(self._ast)
        headings.insert(0, (self._INITIAL_SECTION_NAME, 1))  # Title is not part of AST
        headings = [(h, l) for h, l in headings if h is not None]
        return headings

    @property
    def regexes(self):
        def _regex(header: str, level: int):
            return r"#{{{l}}}\s*{h}[^\S\r\n]*[\n]".format(l=level, h=header.strip())

        return {(h, l): _regex(h, l) for h, l in self.headings}

    @property
    def markdown(self):
        return getattr(self, "_markdown", None)

    @markdown.setter
    def markdown(self, value: str | Path):
        self.fp = value
        self.load()
        return


class Readme(MarkdownLoaderMixin):
    def __init__(self, readme_fp: str | Path):
        self.fp = readme_fp
        self.load()
        return



from functools import partial

from ..functiondict_mixin import FunctionDictMixin


class Fixes:
    @property
    def available_fixes(self):
        return dict(
            stylize_latin_abbreviations=self.stylize_latin_abbreviations,
            no_spaces_before_a_colon=self.no_spaces_before_a_colon,
            avoid_contractions=self.avoid_contractions,
        )

    def stylize_latin_abbreviations(
        self, line: str, verbose: bool | None = None
    ) -> str:
        """
        Automagically add commas after 'i.e.' and 'e.g.'

        Note that such double punctuation (for example: 'e.g.,') is still a matter of active debate.
        Long story short, it almost always happens in American English and tends to be avoided in British English.

        Hence, this function enforces the American convention, not a universal truth.

        However, regional conventions aside, this function also enforces **consistency** across the documentation,
        which is a virtue by itself.

        See [here](https://english.stackexchange.com/questions/16172/should-i-always-use-a-comma-after-e-g-or-i-e)
        for a very informative discussion (which links to various authoratative sources) on this topic.

        Args:
            line (str): Line to be checked and modified if appropriate.
        """
        if verbose is None:
            verbose = self.verbose

        new_line = (
            line.replace("i.e. ", "i.e., ")
            .replace("I.e. ", "I.e., ")
            .replace("e.g. ", "e.g., ")
            .replace("E.g. ", "E.g., ")
        )
        if new_line != line and verbose:
            msg = """
            [STYLE GUIDE] Follow American convention for latin abbreviations; write 'i.e.,' and 'e.g.,'. 
                REPLACED THIS: 
                    {l}
                BY THIS:
                    {n}
            """.format(
                l=line, n=new_line
            )
            print(msg)
        return new_line

    def no_spaces_before_a_colon(self, line: str, verbose: bool | None = None) -> str:
        if verbose is None:
            verbose = self.verbose

        new_line = line.replace(" :", ":")

        if new_line != line and verbose:
            msg = """
            [STYLE GUIDE] Do not leave a space in front of a colon.
                REPLACED THIS: 
                    {l}
                BY THIS:
                    {n}
            """.format(
                l=line, n=new_line
            )
            print(msg)
        return new_line

    def avoid_contractions(self, line: str, verbose: bool | None = None) -> str:
        if verbose is None:
            verbose = self.verbose

        new_line = line.replace("n't", " not").replace("'ll", " will")

        if new_line != line and verbose:
            msg = """
            [STYLE GUIDE] Avoid contractions such as can't, don't, etc.
                REPLACED THIS: 
                    {l}
                BY THIS:
                    {n}
            """.format(
                l=line, n=new_line
            )
            print(msg)
        return new_line


class FixesMixin(Fixes, FunctionDictMixin):
    def do_fixes(self, line, key: str | None = None) -> str:
        """Do all fixes. This may or may not involve transforming the manifest."""
        return self.do_functions(line, key=key, kind="fixes")

    def add_fix(
        self,
        fix: str,
        key: str | None = None,
        **kwargs,
    ):

        return self.add_function(function=fix, key=key, kind="fixes", **kwargs)

    def add_fixes(
        self,
        fixes: list[str],
        key: str | None = None,
        **kwargs,
    ):
        for fix in fixes:
            self.add_fix(fix, key=key, **kwargs)
        return

    def add_all_fixes(self, key: str | None = None, **kwargs):
        """
        Associate all available fixes with a particular key.

        Args:
            key (str | None, optional): Key for this collection of fixes. Defaults to None.
        """
        return self.add_fixes(self.available_fixes, key=key, **kwargs)



# useful fs constants
MF_FP = "/batchx/manifest/manifest.json"
INPUT_FP = "/batchx/input/input.json"
OUTPUT_FP = "/batchx/output/output.json"
DAG_FP = "/batchx/output/dag.svg"
SNAKEMAKE_EXIT_CODE_FP = "/tmp/snakemake_exit_code.json"

# runtime constants
DEFAULT_MEMORY_MB = 4096
DEFAULT_VCPUS = 1



# imports
import json
import os
from datetime import datetime

import click
from batchx_dev.toolbox import Manifest, ManifestImprover, ReadmeImprover

from .end2end import (
    basic_configuration_mfi,
    basic_configuration_rmi,
    make_backup_copy,
    make_initial_readme,
    print_available_improvements,
    print_report_header,
)
from .monkeypatch import Path


# actual function
def run_bx_readme(
    manifest_fp: Path,
    readme_fp: Path,
    backup_dp: Path | None = None,
    manifest_fp_out: Path | None = None,
    readme_fp_out: Path | None = None,
    offline: bool = False,
    bx_env: str = "batchx",
    include_drafts: bool = True,
    verbose: bool = True,
):
    online = not offline

    print_report_header("Initialization", level=1)

    # INIT MFI
    BU_MF_FP = make_backup_copy(manifest_fp, dp=backup_dp)

    if manifest_fp_out is None:
        MF_FP_OUT = manifest_fp

    mfi = ManifestImprover(manifest_fp=manifest_fp, verbose=verbose)

    # INIT RMI
    if not Path(readme_fp).exists():
        make_initial_readme(readme_fp, is_pipeline=mfi.mf.is_pipeline)
    else:
        pass
    BU_RM_FP = make_backup_copy(readme_fp, dp=backup_dp)

    if readme_fp_out is None:
        RM_FP_OUT = readme_fp

    rmi = ReadmeImprover(readme_fp=readme_fp, verbose=verbose)

    if verbose:
        print_report_header("Backups", level=2)
        for e in [BU_MF_FP, BU_RM_FP]:
            print("    - {}".format(e))

    if online and isinstance(getattr(mfi, "steps", None), dict):
        steps_dict = mfi.steps.copy()
    else:
        steps_dict = dict()

    if verbose:
        print_available_improvements(mfi, level=2)
        print_available_improvements(rmi, level=2)

    # MFI
    print_report_header("ManifestImprover", level=1)
    mfi = basic_configuration_mfi(mfi, steps_dict=steps_dict, verbose=verbose)

    # run mfi
    mfi.improve_descriptions_from_name()
    mfi.improve_titles()
    mfi.improve_descriptions()
    mfi.improve_descriptions_from_title()
    mfi.improve_titles_from_name()

    mfi.mf.save(fp=MF_FP_OUT)

    # RMI
    print_report_header("ReadmeImprover", level=1)

    mfi.add_enrichment("fetch_defaults", key="defaults_from_name")
    mfi.improve_defaults_from_name()  # These defaults should not end up on disk

    mfi.add_enrichment("fetch_available", key="available_from_name")
    mfi.improve_available_from_name()  # These defaults should not end up on disk

    rmi = basic_configuration_rmi(
        rmi,
        mfi,
        steps_dict=steps_dict,
        bx_env=bx_env,
        include_drafts=include_drafts,
        flatten=("default", "available"),
        verbose=verbose,
    )

    # run rmi
    rmi.improve_user_defined_sections()
    rmi.generate_platform_defined_sections()
    rmi.improve_platform_defined_sections()
    rmi.draft_user_defined_sections()

    rmi.rm.save(fp=RM_FP_OUT)

    return


# CLI-method
@click.command()
@click.option(
    "--manifest_fp",
    default="manifest.json",
    help="Manifest filepath.",
    type=click.Path(exists=True),
)
@click.option(
    "--readme_fp",
    default="readme.md",
    help="Readme filepath.",
)
@click.option(
    "--backup_dp",
    default=None,
    help="Backup directory path.",
)
@click.option(
    "--manifest_fp_out",
    default=None,
    help="Manifest output filepath.",
)
@click.option(
    "--readme_fp_out",
    default=None,
    help="Readme output filepath.",
)
@click.option(
    "--bx_env",
    default="batchx",
    help="BatchX environment to assume for consistency checks in example section.",
)
@click.option(
    "--offline",
    is_flag=True,
    default=False,
    help="Flag that indicates whether the ReadmeImprover should try and fetch information online.",
)
@click.option(
    "--include_drafts",
    is_flag=True,
    default=True,
    help="Flag that indicates whether the ReadmeImprover should generate drafts for empty sections.",
)
def main(
    manifest_fp: Path,
    readme_fp: Path,
    backup_dp: Path | None = None,
    manifest_fp_out: Path | None = None,
    readme_fp_out: Path | None = None,
    offline: bool = False,
    bx_env: str = "batchx",
    include_drafts: bool = True,
    verbose: bool = True,
):
    return run_bx_readme(
        manifest_fp=manifest_fp,
        readme_fp=readme_fp,
        backup_dp=backup_dp,
        manifest_fp_out=manifest_fp_out,
        readme_fp_out=readme_fp_out,
        offline=offline,
        bx_env=bx_env,
        include_drafts=include_drafts,
        verbose=verbose,
    )


# main clause
if __name__ == "__main__":
    main()



import os
from pathlib import Path

# filesystem
ROOT = Path("/batchx")
TMP = Path("/tmp")
INPUT_DP = ROOT / "input"
OUTPUT_DP = ROOT / "output"
MANIFEST_DP = ROOT / "manifest"
INPUT_JSON_FP = INPUT_DP / "input.json"
OUTPUT_JSON_FP = OUTPUT_DP / "output.json"
MANIFEST_JSON_FP = MANIFEST_DP / "manifest.json"

BX_VCPUS = os.environ.get("BX_VCPUS", None)
BX_MEMORY = os.environ.get("BX_MEMORY", None)

# documentation URLS
URLS = dict(
    BED="https://genome.ucsc.edu/FAQ/FAQformat.html#format1",
    BAM="https://samtools.github.io/hts-specs/SAMv1.pdf",
    BAI="https://samtools.github.io/hts-specs/SAMv1.pdf#subsection.5.2",
    SAM="https://samtools.github.io/hts-specs/SAMv1.pdf",
    VCF="https://genome.ucsc.edu/goldenPath/help/vcf.html",
    GTF="https://genome.ucsc.edu/FAQ/FAQformat.html#format3",
    FAST5="https://medium.com/@shiansu/a-look-at-the-nanopore-fast5-format-f711999e2ff6",
    FASTA="https://en.wikipedia.org/wiki/FASTA_format",
    FASTQ="https://en.wikipedia.org/wiki/FASTQ_format",
    FAI="https://manpages.ubuntu.com/manpages/bionic/man5/faidx.5.html",
    CRAM="https://en.wikipedia.org/wiki/CRAM_(file_format)",
    PILEUP="https://en.wikipedia.org/wiki/Pileup_format",
)

FIXED_DESCRIPTIONS = {
    "schema.input.properties.sample.description": "This pipeline has the following **sample** inputs. These relate to the samples being analyzed. ",
    "schema.input.properties.global.description": "This pipeline has the following **global** inputs. These are files or values that can be reused in several workflows over different samples. ",
    "schema.input.properties.tools.description": "This pipeline has the following **tool** inputs. These are the tools this pipeline orchestrates along with their respective inputs. ",
    "schema.output.properties.dag.description": "SVG graph with the DAG of executed jobs. ",
    "schema.output.properties.tools.description": "This pipeline provides the following outputs, grouped by tool. ",
}

# readme-headers (deprecated!)
README_HEADERS = dict(
    context="# Context",
    inputs="# Inputs",
    required_inputs="## Required inputs",
    optional_inputs="## Optional inputs",
    required_outputs="## Required outputs",
    optional_outputs="## Optional outputs",
    outputs="# Outputs",
    examples="# Examples",
    links="# Links",
    tool_versions="# Tool versions",
    pipeline_overview="# Pipeline overview",
)
